#!/usr/local/bin/python
# -*- coding: utf-8 -*-
"""
The language of positive RIF conditions determines what can appear as a body (the
 if-part) of a rule supported by the basic RIF logic. As explained in Section 
 Overview, RIF's Basic Logic Dialect corresponds to definite Horn rules, and the
  bodies of such rules are conjunctions of atomic formulas without negation.
"""
import itertools
from rdflib import Variable, BNode, URIRef, Literal, Namespace,RDF,RDFS
from rdflib.Literal import _XSD_NS
from rdflib.util import first
from rdflib.Collection import Collection
from rdflib.Graph import ConjunctiveGraph,QuotedGraph,ReadOnlyGraphAggregate, Graph
from rdflib.syntax.NamespaceManager import NamespaceManager

OWL    = Namespace("http://www.w3.org/2002/07/owl#")

def buildUniTerm((s,p,o),newNss=None):
    return Uniterm(p,[s,o],newNss=newNss)

class QNameManager(object):
    def __init__(self,nsDict=None):
        self.nsDict = nsDict and nsDict or {}
        self.nsMgr = NamespaceManager(Graph())
        self.nsMgr.bind('owl','http://www.w3.org/2002/07/owl#')
        self.nsMgr.bind('math','http://www.w3.org/2000/10/swap/math#')
        
    def bind(self,prefix,namespace):
        self.nsMgr.bind(prefix,namespace)

class SetOperator(object):
    def repr(self,operator):
        nafPrefix = self.naf and 'not ' or ''
        if len(self.formulae)==1:
            return nafPrefix+repr(self.formulae[0])
        else:
            return "%s%s( %s )"%(nafPrefix,operator,' '.join([repr(i) for i in self.formulae]))
    def remove(self,item):
        self.formulae.remove(item)
        
    def __len__(self):
        return len(self.formulae)

class Condition(object):
    """
    CONDITION   ::= CONJUNCTION | DISJUNCTION | EXISTENTIAL | ATOMIC
    """
    def isSafeForVariable(self,var):
        """
        A variable, v is safe in a condition formula if and only if ..
        """
        return False
    
    def __iter__(self):
        for f in self.formulae:
            yield f

class And(QNameManager,SetOperator,Condition):
    """
    CONJUNCTION ::= 'And' '(' CONDITION* ')'
    
    >>> And([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
    ...      Uniterm(RDF.type,[OWL.Class,RDFS.Class])])
    And( rdf:Property(rdfs:comment) rdfs:Class(owl:Class) )
    """
    def __init__(self,formulae=None,naf=False):
        self.naf = naf
        self.formulae = formulae and formulae or []
        QNameManager.__init__(self)

    def binds(self, var):
        """
        A variable, v, is bound in a conjunction formula, f = And(c1...cn), n ≥ 1, 
        if and only if, either

        - v is bound in at least one of the conjuncts;

        For now we don't support equality predicates, so we only check the 
        first condition
        
        >>> x=Variable('X')
        >>> y=Variable('Y')
        >>> lit1 = Uniterm(RDF.type,[x,RDFS.Class])
        >>> lit2 = Uniterm(RDF.Property,[y,RDFS.Class])
        >>> conj = And([lit1,lit2])
        >>> conj.binds(Variable('Y'))
        True
        >>> conj.binds(Variable('Z'))
        False
        """
        return first(itertools.ifilter(lambda conj: conj.binds(var),
                                       self.formulae)) is not None
        
    def isSafeForVariable(self,var):
        """
        A variable, v is safe in a condition formula if and only if ..
        
        f is a conjunction, f = And(c1...cn), n ≥ 1, and v is safe in at least 
        one conjunct in f
        
        Since we currently don't support equality predicates, we only check
        the first condition
        
        >>> x=Variable('X')
        >>> y=Variable('Y')
        >>> lit1 = Uniterm(RDF.type,[x,RDFS.Class])
        >>> lit2 = Uniterm(RDF.Property,[y,RDFS.Class])
        >>> conj = And([lit1,lit2])
        >>> conj.isSafeForVariable(y)
        True
                
        """
        return first(itertools.ifilter(lambda conj: conj.isSafeForVariable(var),
                                       self.formulae)) is not None
        
    def n3(self):
        """
        >>> And([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
        ...      Uniterm(RDF.type,[OWL.Class,RDFS.Class])]).n3()
        u'rdfs:comment a rdf:Property .\\n owl:Class a rdfs:Class'
        
        """
#        if not [term for term in self if not isinstance(term,Uniterm)]:
#            g= Graph(namespace_manager = self.nsMgr)
#            g.namespace_manager= self.nsMgr
#            [g.add(term.toRDFTuple()) for term in self]
#            return g.serialize(format='n3')
#        else:
        return u' .\n '.join([i.n3() for i in self])
        
    def __repr__(self):
        return self.repr('And')
    
class Or(QNameManager,SetOperator,Condition):
    """
    DISJUNCTION ::= 'Or' '(' CONDITION* ')'
    
    >>> Or([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
    ...      Uniterm(RDF.type,[OWL.Class,RDFS.Class])])
    Or( rdf:Property(rdfs:comment) rdfs:Class(owl:Class) )
    """
    def __init__(self,formulae=None,naf=False):
        self.naf = naf
        self.formulae = formulae and formulae or []
        QNameManager.__init__(self)

    def binds(self, var):
        """
        A variable, v, is bound in a disjunction formula, if and only if v is 
        bound in every disjunct where it occurs
        
        >>> x=Variable('X')
        >>> y=Variable('Y')
        >>> lit1 = Uniterm(RDF.type,[x,RDFS.Class])
        >>> lit2 = Uniterm(RDF.Property,[y,RDFS.Class])
        >>> conj = And([lit1,lit2])
        >>> disj = Or([conj,lit2])
        >>> disj.binds(y)
        True
        >>> disj.binds(Variable('Z'))
        False
        >>> lit = Uniterm(RDF.type,[OWL.Class,RDFS.Class])
        >>> disj= Or([lit,lit])
        >>> disj.binds(x)
        False
        """
        unboundConjs = list(itertools.takewhile(lambda conj: conj.binds(var),
                                                self.formulae))
        return len(unboundConjs) == len(self.formulae)

    def isSafeForVariable(self,var):
        """
        A variable, v is safe in a condition formula if and only if ..
        
        f is a disjunction, and v is safe in every disjunct;
        """
        unboundConjs = list(itertools.takewhile(
                                    lambda conj: conj.isSafeForVariable(var),
                                    self.formulae))
        return len(unboundConjs) == len(self.formulae)

        
    def __repr__(self):
        return self.repr('Or')

class Exists(Condition):
    """
    EXISTENTIAL ::= 'Exists' Var+ '(' CONDITION ')'
    >>> Exists(formula=Or([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
    ...                    Uniterm(RDF.type,[OWL.Class,RDFS.Class])]),
    ...        declare=[Variable('X'),Variable('Y')])
    Exists ?X ?Y ( Or( rdf:Property(rdfs:comment) rdfs:Class(owl:Class) ) )
    """
    def __init__(self,formula=None,declare=None):
        self.formula = formula
        self.declare = declare and declare or []

    def binds(self, var):
        """
        A variable, v, is bound in an existential formula, 
        Exists v1,...,vn (f'), n ≥ 1, if and only if v is bound in f'
        
        >>> ex=Exists(formula=And([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
        ...                    Uniterm(RDF.type,[Variable('X'),RDFS.Class])]),
        ...        declare=[Variable('X')])
        >>> ex.binds(Variable('X'))
        True
        """
        return self.formula.binds(var)
    
    def isSafeForVariable(self,var):
        """
        A variable, v is safe in a condition formula if and only if ..
        
        f is an existential formula, f = Exists v1,...,vn (f'), n ≥ 1, and 
        v is safe in f' .
        """
        return self.formula.isSafeForVariable(var) 
            
    def __iter__(self):
        for term in self.formula: 
            yield term
    def n3(self):
        """
        """
        return self.formula.n3()
        #return u"@forSome %s %s"%(','.join(self.declare),self.formula.n3())
    
    def __repr__(self):
        return "Exists %s ( %r )"%(' '.join([var.n3() for var in self.declare]),
                                   self.formula )
        
class Atomic(Condition):
    """
    ATOMIC ::= Uniterm | Equal | Member | Subclass (| Frame)
    """
    def binds(self, var):
        """
        A variable, v, is bound in an atomic formula, a, if and only if
        
        - a is neither an equality nor an external predicate, and v occurs as an 
          argument in a;
        - or v is bound in the conjunction formula f = And(a).
        
        Default is False
                
        """
        return False
    
    def __iter__(self):
        yield self

class Equal(QNameManager,Atomic):
    """
    Equal ::= TERM '=' TERM
    TERM ::= Const | Var | Uniterm | 'External' '(' Expr ')'
    
    >>> Equal(RDFS.Resource,OWL.Thing)
    rdfs:Resource =  owl:Thing
    """
    def __init__(self,lhs=None,rhs=None):
        self.lhs = lhs
        self.rhs = rhs
        QNameManager.__init__(self)
        
    def __repr__(self):
        left  = self.nsMgr.qname(self.lhs)
        right = self.nsMgr.qname(self.rhs)
        return "%s =  %s"%(left,right)

def BuildUnitermFromTuple((s,p,o),newNss=None):
    return Uniterm(p,[s,o],newNss)

class Uniterm(QNameManager,Atomic):
    """
    Uniterm ::= Const '(' TERM* ')'
    TERM ::= Const | Var | Uniterm
    
    We restrict to binary predicates (RDF triples)
    
    >>> Uniterm(RDF.type,[RDFS.comment,RDF.Property])
    rdf:Property(rdfs:comment)
    """
    def __init__(self,op,arg=None,newNss=None,naf=False):
        self.naf = naf
        self.op = op
        self.arg = arg and arg or []
        QNameManager.__init__(self)
        if newNss is not None:
            newNss = newNss.items() if isinstance(newNss,dict) else newNss
            for k,v in newNss:
                self.nsMgr.bind(k,v)
        self._hash=hash(reduce(lambda x,y:str(x)+str(y),
            len(self.arg)==2 and self.toRDFTuple() or [self.op]+self.arg))

    def binds(self, var):
        """
        A variable, v, is bound in an atomic formula, a, if and only if
        
        - a is neither an equality nor an external predicate, and v occurs as an 
          argument in a;
        - or v is bound in the conjunction formula f = And(a).
        
        Default is False
        
        >>> x = Variable('X')
        >>> lit = Uniterm(RDF.type,[RDFS.comment,x])
        >>> lit.binds(Variable('Z'))
        False
        >>> lit.binds(x)
        False
        >>> Uniterm(RDF.type,[x,RDFS.Class]).binds(x)
        True
        
        """
        if self.op == RDF.type:
            arg0,arg1 = self.arg
            return var == arg0            
        else:
            return var in self.arg
        
    def isSafeForVariable(self,var):
        """
        A variable, v is safe in a condition formula if and only if ..
        
        f is an atomic formula and f is not an equality formula in which both 
        terms are variables, and v occurs in f;
        """
        return self.binds(var)
        
    def renameVariables(self,varMapping):
        if self.op == RDF.type:
            self.arg[0]=varMapping.get(self.arg[0],self.arg[0])
        else:
            self.arg[0]=varMapping.get(self.arg[0],self.arg[0])
            self.arg[1]=varMapping.get(self.arg[1],self.arg[1])
        #Recalculate the hash after modification
        self._hash=hash(reduce(lambda x,y:str(x)+str(y),
            len(self.arg)==2 and self.toRDFTuple() or [self.op]+self.arg))
            
            
    def _get_terms(self):
        """
        Class attribute that returns all the terms of the literal as a lists
        >>> x = Variable('X')
        >>> lit = Uniterm(RDF.type,[RDFS.comment,x])
        >>> lit.terms
        [rdflib.URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'), rdflib.URIRef('http://www.w3.org/2000/01/rdf-schema#comment'), ?X]
        """
        return [self.op]+self.arg
            
    terms = property(_get_terms)
            
    def getVarMapping(self,otherLit, reverse=False):
        """
        Takes another Uniterm and in every case where the corresponding term 
        for both literals are different variables, we map from the variable
        for *this* uniterm to the corresponding variable of the other.
        The mapping will go in the other direction if the reverse
        keyword is True
          
        >>> x = Variable('X')
        >>> y = Variable('Y')
        >>> lit1 = Uniterm(RDF.type,[RDFS.comment,x])
        >>> lit2 = Uniterm(RDF.type,[RDFS.comment,y])
        >>> lit1.getVarMapping(lit2)[x] == y
        True
        >>> lit1.getVarMapping(lit2,True)[y] == x
        True
        """
        map = {}
        if isinstance(self.op,Variable) and isinstance(otherLit.op,Variable) and \
            self.op != otherLit.op:
            if reverse:
                map[otherLit.op] = self.op 
            else:
                map[self.op] = otherLit.op
        if isinstance(self.arg[0],Variable) and isinstance(otherLit.arg[0],Variable) \
            and self.arg[0] != otherLit.arg[0]:
            if reverse:
                map[otherLit.arg[0]] = self.arg[0]
            else:
                map[self.arg[0]] = otherLit.arg[0]
        if isinstance(self.arg[1],Variable) and isinstance(otherLit.arg[1],Variable) and \
            self.arg[1] != otherLit.arg[1]:
            if reverse:
                map[otherLit.arg[1]] = self.arg[1] 
            else:
                map[self.arg[1]] = otherLit.arg[1]
        return map

    def applicableMapping(self,mapping):
        """
        Can the given mapping (presumably from variables to terms) be applied?
        """
        return bool(set(mapping).intersection([self.op]+self.arg))        

    def ground(self,varMapping):
        appliedKeys = set([self.op]+self.arg).intersection(varMapping.keys())
        self.op    =varMapping.get(self.op,self.op)
        self.arg[0]=varMapping.get(self.arg[0],self.arg[0])
        self.arg[1]=varMapping.get(self.arg[1],self.arg[1])
        #Recalculate the hash after modification
        self._hash=hash(reduce(lambda x,y:str(x)+str(y),
            len(self.arg)==2 and self.toRDFTuple() or [self.op]+self.arg))        
        return appliedKeys
            
    def isGround(self):
        for term in [self.op]+self.arg:
            if isinstance(term,Variable):
                return False
        return True
                
    def __hash__(self):
        return self._hash
    
    def __eq__(self,other):
        return hash(self) == hash(other)                       
        
    def renderTermAsN3(self,term):
        if term == RDF.type:
            return 'a'
        elif isinstance(term, (BNode,Literal,Variable)):
            return term.n3()
        else:
            return self.nsMgr.qname(term)
        
    def n3(self):
        """
        Serialize as N3 (using available namespace managers)
        
        >>> Uniterm(RDF.type,[RDFS.comment,RDF.Property]).n3()
        u'rdfs:comment a rdf:Property'

        """
        return ' '.join([ self.renderTermAsN3(term) 
                         for term in [self.arg[0],self.op,self.arg[1]]])
        
    def toRDFTuple(self):
        subject,_object = self.arg
        return (subject,self.op,_object)
        
    def collapseName(self,val):
        try:
            rt = self.nsMgr.qname(val)
            if len(rt.split(':')[0])>1 and rt[0]=='_':
                return u':'+rt.split(':')[-1]
            else:
                return rt
                
        except:
            for prefix,uri in self.nsMgr.namespaces():
                if val.startswith(uri):
                    return u'%s:%s'%(prefix,val.split(uri)[-1])
            return val
        
    def normalizeTerm(self,term):
        if isinstance(term,Literal):
            if term.datatype == _XSD_NS.integer:
                return unicode(term)
            else:
                return term.n3()
        else:
            return isinstance(term,Variable) and term.n3() or \
                   self.collapseName(term)        
   
    def getArity(self):
        return 1 if self.op == RDF.type else 2
        
    arity = property(getArity)
   
    def setOperator(self,newOp):
        if self.op == RDF.type:
            self.arg[-1] = newOp
        else:
            self.op = newOp
        
    def isSecondOrder(self):
        if self.op == RDF.type:
            return isinstance(self.arg[-1],Variable)
        else:
            return isinstance(self.op,Variable)
        
    def __repr__(self):
        negPrefix = self.naf and 'not ' or ''
        if self.op == RDF.type:
            arg0,arg1 = self.arg
            return "%s%s(%s)"%(negPrefix,self.normalizeTerm(arg1),self.normalizeTerm(arg0))
        else:
            return "%s%s(%s)"%(negPrefix,self.normalizeTerm(self.op),
                                ' '.join([self.normalizeTerm(i) for i in self.arg]))
            
class PredicateExtentFactory(object):
    """
    Creates an object which when indexed returns
    a Uniterm with the 'registered' symbol and
    two-tuple argument
    
    >>> from rdflib import Namespace, URIRef
    >>> EX_NS = Namespace('http://example.com/')
    >>> ns = {'ex':EX_NS}
    >>> somePredFactory = PredicateExtentFactory(EX_NS.somePredicate,newNss=ns)
    >>> somePredFactory[(EX_NS.individual1,EX_NS.individual2)]
    ex:somePredicate(ex:individual1 ex:individual2)
    >>> somePred2Factory = PredicateExtentFactory(EX_NS.somePredicate,binary=False,newNss=ns)
    >>> somePred2Factory[EX_NS.individual1]
    ex:somePredicate(ex:individual1)
    
    """
    def __init__(self,predicateSymbol,binary=True,newNss=None):
        self.predicateSymbol = predicateSymbol
        self.binary = binary
        self.newNss = newNss
        
    def term(self, name):
        return Class(URIRef(self + name))

    def __getitem__(self, args):
        if self.binary:
            arg1,arg2=args
            return Uniterm(self.predicateSymbol,[arg1,arg2],newNss=self.newNss)
        else:
            return Uniterm(RDF.type,[args,self.predicateSymbol],newNss=self.newNss)
                        
class ExternalFunction(Uniterm):
    """
    An External(ATOMIC) is a call to an externally defined predicate, equality, 
    membership, subclassing, or frame. Likewise, External(Expr) is a call to an 
    externally defined function.
    >>> ExternalFunction(Uniterm(URIRef('http://www.w3.org/2000/10/swap/math#greaterThan'),[Variable('VAL'),Literal(2)]))
    math:greaterThan(?VAL "2"^^<http://www.w3.org/2001/XMLSchema#integer>)
    """
    def __init__(self,builtin,newNss=None):
        from FuXi.Rete.RuleStore import N3RuleStore,N3Builtin
        self.builtin= builtin
        if isinstance(builtin,N3Builtin):
            Uniterm.__init__(self,builtin.uri,[builtin.argument,builtin.result])
        else:
            Uniterm.__init__(self,builtin.op,builtin.arg)
        QNameManager.__init__(self)
        if newNss is not None:
            newNss = isinstance(newNss,dict) and newNss.items() or newNss
            for k,v in newNss:
                self.nsMgr.bind(k,v)
            
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
"""
BackwardFixpointProcedure.py

.. A sound and complete query answering method for recursive databases
based on meta-interpretation called Backward Fixpoint Procedure ..

Uses RETE-UL as the RIF PRD implementation of
a meta-interpreter of an adorned ruleset that builds large, conjunctive (BGPs) SPARQL queries.

Facts are only generated in a bottom up evaluation of the interpreter if a query has been issued
for that fact or if an appropriate sub-query has been generated. Sub-queries for rule bodies are generated 
if a sub-query for the corresponding rule head already exists. Sub-queries for conjuncts
are generated from sub-queries of conjunctions they appear in 

Evaluate condition and ACTION:

Evaluate consults the already generated facts, and may take a single
atom or a conjunction as its argument, returning true if all of the conjuncts have
already been generated.

"""

import sys, copy, os, unittest, copy
from pprint import pprint
from rdflib import URIRef, RDF, RDFS, Namespace, Variable, Literal, URIRef
from rdflib.util import first
from rdflib.Graph import ReadOnlyGraphAggregate
from FuXi.SPARQL import EDBQuery
from FuXi.Rete.SidewaysInformationPassing import GetArgs, GetVariables, SIPRepresentation
from FuXi.Horn.PositiveConditions import *
from FuXi.Rete.SidewaysInformationPassing import iterCondition, GetOp
from FuXi.Rete.BetaNode import ReteMemory, BetaNode, RIGHT_MEMORY, LEFT_MEMORY, collectVariables, PartialInstanciation
from FuXi.Rete.AlphaNode import AlphaNode, ReteToken
from FuXi.Rete.Network import ReteNetwork, HashablePatternList, InferredGoal
from FuXi.Rete.Proof import MakeImmutableDict
from FuXi.DLP import breadth_first
from FuXi.Rete.Magic import AdornedRule, AdornedUniTerm
from FuXi.Rete.Util import generateTokenSet, selective_memoize
from FuXi.Horn.HornRules import extractVariables, Clause
from FuXi.Rete.RuleStore import N3Builtin
from cStringIO import StringIO

BFP_NS   = Namespace('http://dx.doi.org/10.1016/0169-023X(90)90017-8#')
BFP_RULE = Namespace('http://code.google.com/p/python-dlp/wiki/BFPSpecializedRule#')
HIGHER_ORDER_QUERY = BFP_RULE.SecondOrderPredicateQuery

class EvaluateConjunctiveQueryMemory(ReteMemory):
    """
    The extension of the evaluate predicate for a particular specialized rule
    
    "Whenever a new WME is filtered through the alpha network and reaches an alpha memory, we
    simply add it to the list of other WMEs in that memory, and inform each of the attached join
    nodes"
    
    A beta memory node stores a list of the tokens it contains, plus a list of its children (other
    nodes in the beta part of the network). Before we give its data structure, though, recall that
    we were going to do our procedure calls for left and right activations through a switch or case
    statement or a jumptable indexed according to the type of node being activated. Thus, given
    a (pointer to a) node, we need to be able to determine its type. This is straightforward if we
    use variant records to represent nodes. (A variant record is a record that can contain any one
    of several dierent sets of elds.) Each node in the beta part of the net will be represented by
    a rete-node structure:
    
    Whenever a beta memory is informed of a new match (consisting of an existing token and some
    WME), we build a token, add it to the list in the beta memory, and inform each of the beta
    memory's children:
    """
    def __init__(self,betaNode,memoryPos,_filter=None):
        super(EvaluateConjunctiveQueryMemory, self).__init__(betaNode,memoryPos,_filter)

    # def union(self, other):
    #     """Return the union of two sets as a new set.
    # 
    #     (I.e. all elements that are in either set.)
    #     """
    #     result = ReteMemory(self.successor,self.position)
    #     result._update(other)
    #     return result    

    def __repr__(self):
        return "<Evaluate Memory: %s item(s)>"%(len(self))

    # def addToken(self,token,debug=False): pass

    # def reset(self):
    #     self.clear()
    #     self.substitutionDict = {}

class MalformedQeryPredicate(Exception):
    """An exception raised when a malformed quer predicate is created"""
    def __init__(self, msg):
        super(MalformedQeryPredicate, self).__init__(msg)


# coroutine.py
#
# A decorator function that takes care of starting a coroutine
# automatically on call.

def coroutine(func):
    def start(*args,**kwargs):
        cr = func(*args,**kwargs)
        cr.next()
        return cr
    return start
                
class GoalSolutionAction(object):
    def __init__(self, bfp, varMap):
        self.bfp = bfp
        self.varMap = varMap
        self.solutionSet = set()
        # self.goal = BuildUnitermFromTuple(goal)
    
    def __repr__(self):
        stream = StringIO()
        pprint(list(self.solutionSet),stream)
        return stream.getvalue()
        
    def __call__(self, tNode, inferredTriple, token, binding, debug):
        """
        Called when the BFP triggers a p-node associated with a goal
        , storing the solutions for later retrieval
        """
        self.bfp.goalSolutions.add(
            MakeImmutableDict(
                dict(
                    [(self.varMap[key],binding[key]) 
                        for key in binding
                          if key in self.varMap])))
                
class EvaluateExecution(object):
    """Handles the inference of evaluate literals in BFP"""
    def __init__(self, (ruleNo,bodyIdx),bfp,termNodes):
        self.ruleNo        = ruleNo
        self.bodyIdx       = bodyIdx
        self.bfp           = bfp
        self.termNodes     = termNodes
        for termNode in self.termNodes:
            assert [ (s,p,o) 
                        for s,p,o in termNode.consequent 
                            if p == BFP_NS.evaluate and \
                               s == BFP_RULE[str(self.ruleNo)] and \
                               o == Literal(self.bodyIdx) ],"%s %s"%(self,termNode)

    
    def __call__(self, tNode, inferredTriple, token, binding, debug):
        """
        Called when an evaluate literal is inferred and
        given the relevant bindings
        
        Add entailed evaluate bindings (as high-arity predicates) 
        directly into RETE-UL beta node memories in a circular fashion 
        propagating their sucessor        
        """
        for s,p,o in tNode.consequent:
            if p == BFP_NS.evaluate:
                for memory in self.bfp.evalHash[(self.ruleNo,self.bodyIdx)]:
                    for bindings in token.bindings:
                        memory.addToken(token,debug)
                        if memory.position == LEFT_MEMORY:
                            memory.successor.propagate(memory.position,debug,token)
                        else:
                            memory.successor.propagate(None,debug,token)

    def __repr__(self):
        return "Evaluate(%s %s) feeds %s memories"%(
                    self.ruleNo,
                    self.bodyIdx,
                    len(self.bfp.memories))

class QueryExecution(object):
    """
    Called when an evaluate literal is inferred and
    given the relevant bindings
    """
    def __init__(self, bfp, queryLiteral):
        self.factGraph = bfp.factGraph
        self.bfp = bfp
        self.queryLiteral = queryLiteral
        self.bfp.firedEDBQueries = set()

    def __call__(self, tNode, inferredTriple, token, binding, debug=False):
        """
        Called when a (EDB) query literal is inferred with
        given bindings.
        """
        assert len(tNode.consequent) == 1
        key = (self.queryLiteral,tNode,token)
        if key not in self.bfp.firedEDBQueries:
            self.bfp.firedEDBQueries.add(key)
            for binding in token.bindings:
                #For each mapping that unifies with theory
                _qLit = copy.deepcopy(self.queryLiteral)
                _bindings = dict([(k,v) for k,v in binding.items() 
                                    if v != OPEN_QUERY_VARIABLE])
                _qLit.ground(_bindings)
                _vars = list(GetVariables(_qLit,secondOrder=True))
                closure = ReadOnlyGraphAggregate([self.factGraph,
                                                  self.bfp.metaInterpNetwork.inferredFacts])
                closure.templateMap = self.factGraph.templateMap
                if self.bfp.debug:
                    print "Query triggered for ", tNode.clause
                edbQuery = EDBQuery([_qLit],closure,_vars,_bindings)
                self.bfp.edbQueries.add(edbQuery)
                queryStr,rt = edbQuery.evaluate(self.bfp.debug)
                if not isinstance(rt,bool):
                    #Non-ground, single atom CQ
                    for ans in rt:
                        self.handleQueryAnswer(_qLit,debug,ans)
                elif rt:
                    self.handleQueryAnswer(_qLit,debug)

    def handleQueryAnswer(self,literal,debug,bindings=None):
        edbResult = copy.deepcopy(literal)
        if bindings:
            edbResult.ground(bindings)    
        inferredToken=ReteToken(edbResult.toRDFTuple(),debug=debug)
        if inferredToken not in self.bfp.metaInterpNetwork.workingMemory:                    
            if self.bfp.debug or debug:
                print "\tAnswer to BFP triggered query %s : %s"%(edbResult,bindings)
            self.bfp.metaInterpNetwork.addWME(inferredToken)                    

    def __repr__(self):
        return "QueryExecution%s"%('( against EDB: %s )'%self.queryLiteral)
        
def SetupEvaluationBetaNode(existingBetaNode,rule,network):
    """
    Take a BetaNode (and a BFP rule) that joins values from an evaluate condition
    with other conditions and replace the alpha node (and memory) used
    to represent the condition with a pass-thru beta with no parent nodes
    but whose right memory will be used to add bindings instanciated
    from evaluate assertions in the BFP algorithm
    
      Rete Network
      ------------

       ...       
         \     memory <-- evalAlphaNode
          \   /    
         existingBetaNode
            
    ...     
      \   evalMemory <-- evaluate(ruleNo,bodyPos,vars)
       \   /    
      existingBetaNode
    """                
    #Delete the existing alpha node (and memory) for the evaluate condition

    newMem = EvaluateConjunctiveQueryMemory(existingBetaNode,RIGHT_MEMORY)
    existingBetaNode.memories[RIGHT_MEMORY] = newMem
    evalAlphaNode = existingBetaNode.rightNode
    network.alphaPatternHash[evalAlphaNode.alphaNetworkHash()][
        evalAlphaNode.alphaNetworkHash(groundTermHash=True)].remove(evalAlphaNode)
    network.alphaNodes.remove(evalAlphaNode)
    for mem in evalAlphaNode.descendentMemory:
        del mem
    pattern = HashablePatternList([evalAlphaNode.triplePattern])
    if pattern  in network.nodes:
        del network.nodes[pattern]
    del evalAlphaNode    
    existingBetaNode.rightNode = None
    
    #The common variables are those in the original rule intersected
    #with those in the left node of the successor
    existingBetaNode.rightVariables = set(rule.declare)
    existingBetaNode.commonVariables = [
            leftVar for leftVar in existingBetaNode.leftVariables
                        if leftVar in existingBetaNode.rightVariables]
    return newMem

        
def NoopCallbackFn(termNode,inferredTriple,tokens,debug=False):
    pass        
                     
OPEN_QUERY_VARIABLE = BFP_NS.NonDistinguishedVariable                     
                                        
class BackwardFixpointProcedure(object):
    """
    Uses RETE-UL as a production rule system implementation of
    a meta-interpreter of an adorned RIF Core ruleset that builds solves conjunctive (BGPs) 
    SPARQL queries.

    Facts are only generated in a bottom up evaluation of the interpreter if a 
    query has been issued for that fact or if an appropriate sub-query has been generated. 
    Sub-queries for rule bodies (conditions) are generated if a sub-query for 
    the corresponding rule head already exists. Sub-queries for conjuncts are 
    generated from sub-queries of conjunctions they appear in (queries are collected).
    """
    def __init__(self, 
                factGraph, 
                network, 
                derivedPredicates, 
                goal,
                sipCollection = [],
                hybridPredicates = None,
                debug = False):
        self.debug = debug
        self.queryPredicates = set()
        self.sipCollection = sipCollection
        self.goal = BuildUnitermFromTuple(goal)
        self.factGraph = factGraph
        self.rules = list(factGraph.adornedProgram)
        self.discardedRules= set()
        self.ruleLabels = {}
        self.bfpLookup = {}
        self.actionHash = {}
        self.namespaces = {
            u'bfp'  : BFP_NS, 
            u'rule' : BFP_RULE 
        }
        self.metaInterpNetwork = network
        self.derivedPredicates = set(derivedPredicates) if \
           isinstance(derivedPredicates,list) else derivedPredicates
        self.hybridPredicates = hybridPredicates        
        self.edbQueries = set()
        self.goalSolutions = set()

    def answers(
        self,
        debug = False,
        solutionCallback = NoopCallbackFn):
        """
        Takes a conjunctive query, a sip collection
        and initiates the meta-interpreter for a given
        goal (at a time), propagating evaluate procedures
        explicitely if no bindings are given from the query
        to trigger subsequent subqueries via EDB predicates
        
        @TODO:
        Add a PRD externally defined action to the 
        production of rules that produce answers
        for the query predicate.
        The action is a user specified callback that can be used
        to signal InferredGoal and halt RETE/UL evaluation prematurely
        otherwise, it is left to reach a stable state and the 
        answers collected along the way are added and returned
                
        """
        solutions = []
        
        queryOp = GetOp(self.goal)
        if self.goal.isGround():
            #Mark ground goal so, production rule engine
            #halts when goal is inferred
            self.metaInterpNetwork.goal = self.goal.toRDFTuple()
    
        adornment = ['f' if isinstance(v,Variable) else 'b' 
                        for v in GetArgs(self.goal,
                                         secondOrder=True)]
        adornment = reduce(lambda x,y:x+y,adornment)
        adornedQuery = AdornedUniTerm(self.goal,adornment)
        bfpTopQuery = self.makeDerivedQueryPredicate(adornedQuery)
        if debug:
            print >>sys.stderr, "Asserting initial BFP query ", bfpTopQuery
            
        assert bfpTopQuery.isGround()
        #Add BFP query atom to working memory, triggering procedure
        try:
            self.metaInterpNetwork.feedFactsToAdd(
                                generateTokenSet(
                                    [bfpTopQuery.toRDFTuple()],
#                                        debugTriples=[bfpTopQuery.toRDFTuple()]
#                                            if debug else []
                                    ))
        except InferredGoal:
            if debug:
                print >>sys.stderr, "Reached ground goal. Terminated BFP!"
            return True
        else:
            if self.goal.isGround():
                #Ground goal, but didn't trigger it, response must be negative
                return False

    def createTopDownReteNetwork(self,debug=False):
        """
        Uses the specialized BFP meta-interpretation rules to build a RETE-UL decision
        network that is modified to support the propagation of bindings from the evaluate
        predicates into a supplimental magic set sip strategy and the generation of subqueries.
        The end result is a bottom-up simulation of SLD resolution with complete, sound, and safe 
        memoization in the face of recursion
        """
        for rule in set(self.specializeAdornedRuleset(debug)):
            self.metaInterpNetwork.buildNetworkFromClause(rule)
#        if debug:
#            sortedBFPRules =[
#                str("%s : %s")%(key,self.bfpLookup[key])
#                    for key in sorted(self.bfpLookup,
#                                      key=lambda items: str(items[1])+items[0])]
#            pprint(sortedBFPRules)
            
        self.evalHash   = {}
        self.actionHash = {}
        evalVars        = {}
        self.productions= {}
        for idx,rule in enumerate(self.rules):
            if rule in self.discardedRules:
                continue                

            label = BFP_RULE[str(idx+1)]
            conjunctLength = len(list(iterCondition(rule.formula.body)))

            #Rule a^k
            #p(x0,...,xn) :- And(query_p(x0,...,xn) evaluate(ruleNo,n,X))
            currentPattern = HashablePatternList([(BFP_RULE[str(idx+1)],
                                                   BFP_NS.evaluate,
                                                   Literal(conjunctLength))])
            assert rule.declare
            #Find alpha node associated with evaluate condition
            node = self.metaInterpNetwork.nodes[currentPattern]
            #evaluate(k,n,X) is a condition in only 1 bfp rule
            assert len(node.descendentBetaNodes) == 1
            bNode = first(node.descendentBetaNodes)
            assert bNode.leftNode.aPassThru
            assert len(bNode.consequent) == 1
            newEvalMemory = SetupEvaluationBetaNode(bNode,rule,self.metaInterpNetwork)
            self.evalHash.setdefault((idx+1,conjunctLength),[]).append(newEvalMemory)
            
            if GetOp(rule.formula.head) == GetOp(self.goal):
                #This rule matches a goal, add a solution collecting action
                goalSolutionAction = GoalSolutionAction(
                        self,
                        rule.formula.head.getVarMapping(self.goal))                
                bNode.executeActions[rule.formula.head.toRDFTuple()] = (
                        False,
                        goalSolutionAction)

            self.productions.setdefault(GetOp(rule.formula.head),[]).append((idx,bNode))

            #Rule b^k
            #evaluate(ruleNo,0,X) :- query_p(x0,...,xn)
            _rule = self.bfpLookup[('b',idx+1)]
            #alpha node associated with query predicate for head of original rule
            _bodyAlphaNode = self.metaInterpNetwork.nodes[
                    HashablePatternList([_rule.formula.body.toRDFTuple()])]
                    
            assert len(_bodyAlphaNode.descendentBetaNodes)==1
            tNode = first(_bodyAlphaNode.descendentBetaNodes)
            self.actionHash.setdefault((idx+1,0),set()).add(tNode)

            #V_{0} = vars(query_p(..))
            headQueryPred = list(iterCondition(_rule.formula.body))[0] 
            try:
                evalVars[(idx+1,0)] = list(headQueryPred.getDistinguishedVariables(True))
            except IndexError:
                self.discardedRules.add(rule)
                continue
            
            for bodyIdx,bodyLiteral in enumerate(iterCondition(rule.formula.body)):
                _ruleC = self.bfpLookup[('c',idx+1,bodyIdx+1)]
                _ruleD = self.bfpLookup[('d',idx+1,bodyIdx+1)]
                pattern = HashablePatternList(
                                        [(BFP_RULE[str(idx+1)],
                                          BFP_NS.evaluate,
                                          Literal(bodyIdx))])
                pattern2 = HashablePatternList(
                                      [None,
                                      (BFP_RULE[str(idx+1)],
                                       BFP_NS.evaluate,
                                       Literal(bodyIdx)),
                                      bodyLiteral.toRDFTuple()])
                aNodeDk = self.metaInterpNetwork.nodes[pattern]
                termNodeCk = self.metaInterpNetwork.nodes[pattern2]

                #Rule d^k
                #query_Literal(x0,...,xj) :- evaluate(ruleNo,j,X)
                assert len(aNodeDk.descendentBetaNodes)==1
                tNode = first(aNodeDk.descendentBetaNodes)
                assert len(tNode.consequent)==1
                if isinstance(bodyLiteral,N3Builtin):
                    self.metaInterpNetwork.alphaNodes.remove(aNodeDk)
                    evalTerm = (BFP_RULE[str(idx+1)],BFP_NS.evaluate,Literal(bodyIdx))
                    token = ReteToken(evalTerm)
                    self.metaInterpNetwork.alphaPatternHash[aNodeDk.alphaNetworkHash()][
                        aNodeDk.alphaNetworkHash(groundTermHash=True)].remove(aNodeDk)
                    del aNodeDk
                    #We bypass d^k, so when evaluate(ruleNo,j,X) is inferred
                    #it is added to left memory of pNode associated with c^k
                    self.evalHash.setdefault((idx+1,bodyIdx),[]).append(termNodeCk.memories[LEFT_MEMORY])
                else:
                    newEvalMemory = SetupEvaluationBetaNode(tNode,rule,self.metaInterpNetwork)
                    self.evalHash.setdefault((idx+1,bodyIdx),[]).append(newEvalMemory)

                    isBase = bodyLiteral.adornment is None if \
                                isinstance(bodyLiteral,AdornedUniTerm) else True
                                
                    if self.hybridPredicates is not None and \
                       GetOp(bodyLiteral) in self.hybridPredicates:
                       isBase = True             
                    if isBase:
                        matchingTriple = first(tNode.consequent)
                        newAction=QueryExecution(
                                self,
                                bodyLiteral)
                        #If the body predicate is a 2nd order predicate, we don't infer the
                        #second order query predicate (since it will trigger a query into
                        #the EDB and thus there is no need to trigger subsequent rules)
                        tNode.executeActions[matchingTriple] = (
                                isinstance(GetOp(bodyLiteral),Variable),
                                newAction)
                #Rule c^k
                #evaluate(ruleNo,j+1,X) :- And(evaluate(ruleNo,j,X) Literal(x0,...,xj))
                self.actionHash.setdefault((idx+1,bodyIdx+1),set()).add(termNodeCk)
                
                termNodeCk.leftVariables = set(rule.declare)
                termNodeCk.commonVariables = list(collectVariables(termNodeCk.rightNode))
                
                #V_{j} = V_{j-1} UNION vars(Literal(..)) where j <> 0
                evalVars[(idx+1,bodyIdx+1)] = list(GetVariables(bodyLiteral,secondOrder=True)) +\
                                                   evalVars[(idx+1,bodyIdx)]
        
        for (ruleNo,bodyIdx),tNodes in self.actionHash.items():
            executeAction = EvaluateExecution((ruleNo,bodyIdx),
                                              self,
                                              tNodes)
            evaluationStmt = (BFP_RULE[str(ruleNo)],BFP_NS.evaluate,Literal(bodyIdx))
            for tNode in tNodes:
                tNode.executeActions[evaluationStmt] = (True,executeAction)
                # executeAction = EvaluateExecution(evalHash,(idx+1,bodyIdx+1),self,termNodeCk)
                # assert len(termNodeCk.consequent)==1
                # termNodeCk.executeAction = (None,executeAction)

        #Fix join variables for BetaNodes involving evaluate predicates
        for idx,rule in enumerate(self.rules):
            if rule in self.discardedRules:
                continue                

            #Rule a^k
            #p(x0,...,xn) :- And(query_p(x0,...,xn) evaluate(ruleNo,n,X))
            #Join vars = vars(query_p) AND V_{n}
            headQueryPred = self.bfpLookup[('b',idx+1)].formula.body
            ruleBodyLen = len(list(iterCondition(rule.formula.body)))
            termNode  = first(self.evalHash[idx+1,ruleBodyLen]).successor
            termNode.commonVariables = list(set(evalVars[(idx+1,ruleBodyLen)]
                                       ).intersection(
                                           GetVariables(headQueryPred,
                                                        secondOrder=True)))
            for bodyIdx,bodyLiteral in enumerate(iterCondition(rule.formula.body)):
                #Rule c^k
                #evaluate(ruleNo,j+1,X) :- And(evaluate(ruleNo,j,X) Literal(x0,...,xj))
                #Join vars = vars(Literal) AND V_{j}
                termNode2 = self.actionHash[idx+1,bodyIdx+1]
                assert len(termNode2)==1
                termNode2 = first(termNode2)
                termNode2.commonVariables = list(set(evalVars[(idx+1,bodyIdx)]
                                           ).intersection(
                                               GetVariables(bodyLiteral,
                                                            secondOrder=True)))                                                            
    def makeDerivedQueryPredicate(self,predicate):
        if isinstance(predicate,AdornedUniTerm):
            newAdornedPred=BFPQueryTerm(predicate,predicate.adornment)
        elif isinstance(predicate,N3Builtin):
            newAdornedPred=BFPQueryTerm(predicate,builtin=predicate)
        else:
            newAdornedPred=BFPQueryTerm(predicate,None)
        if isinstance(newAdornedPred,Uniterm):
            if isinstance(GetOp(newAdornedPred),Variable):
                newAdornedPred.setOperator(HIGHER_ORDER_QUERY)
            newAdornedPred.finalize()
        self.queryPredicates.add(newAdornedPred)
        return newAdornedPred

    def specializeAdornedRuleset(self,debug=False):
        """
        Specialization is applied to the BFP meta-interpreter with respect to the
        rules of the object program. For each rule of the meta-interpreter
        that includes a premise referring to a rule of the object program, one
        specialized version is created for each rule of the object program.
        
        """
        rules = set()
        for idx,rule in enumerate(self.rules):
            label = BFP_RULE[str(idx+1)]
            ruleBodyLen = len(list(iterCondition(rule.formula.body)))
            
            # if rule.isSecondOrder():
            #     if debug:
            #         print "Skipping second order rule (%s): %s"%(idx+1,rule)
            #     continue
            if debug:
                print "\t%s. %s"%(idx+1,rule)
                for _sip in SIPRepresentation(rule.sip):
                    print "\t\t",_sip
                
            newRule1 =self.rule1(rule,label,ruleBodyLen)
            self.bfpLookup[('a',idx+1)] = newRule1
            rules.add(newRule1)
            newRule2=self.rule2(rule,label,ruleBodyLen)
            self.bfpLookup[('b',idx+1)] = newRule2
            rules.add(newRule2)
            for bodyIdx,bodyLiteral in enumerate(iterCondition(rule.formula.body)):
                evaluateTerm = Uniterm(BFP_NS.evaluate,[label,Literal(bodyIdx+1)],
                                       newNss=self.namespaces)
                priorEvaluateTerm = Uniterm(BFP_NS.evaluate,[label,Literal(bodyIdx)],
                                       newNss=self.namespaces)
                #evaluate(ruleNo,j+1,X) :- And(evaluate(ruleNo,j,X) Literal(x0,...,xj))
                newRule = self.makeAdornedRule(
                        And([priorEvaluateTerm,bodyLiteral]),
                             evaluateTerm)
                self.bfpLookup[('c',idx+1,bodyIdx+1)] = newRule
                rules.add(newRule)
                #query_Literal(x0,...,xj) :- evaluate(ruleNo,j,X)
                #OpenQuery(query_Literal)
                newRule = self.makeAdornedRule(
                            priorEvaluateTerm,
                            self.makeDerivedQueryPredicate(bodyLiteral))
                self.bfpLookup[('d',idx+1,bodyIdx+1)] = newRule
                rules.add(newRule)
        return rules


    def makeAdornedRule(self,body,head):
        allVars = set()
        #first we identify body variables
        bodyVars = set(reduce(lambda x,y:x+y,
                              [ list(extractVariables(i,existential=False))
                                        for i in iterCondition(body) ]))
        #then we identify head variables
        headVars = set(reduce(lambda x,y:x+y,
                              [ list(extractVariables(i,existential=False))
                                        for i in iterCondition(head) ]))


        return AdornedRule(Clause(body,head),declare=allVars)


    def rule1(self,rule,label,bodyLen):
        """
        'Facts are only generated in a bottom up evaluation of the interpreter if a query has been issued
         for that fact or if an appropriate sub-query has been generated by the metainterpreter
         itself.'
         
        a^{k} 
         
        p(x0,...,xn) :- And(query_p(x0,...,xn) evaluate(ruleNo,n,X))
                            OpenQuery(query_p)
        
        If there are no bindings posed with the query, then OpenQuery(query_p) 
        is used instead of query_p(x0,...,xn), indicating that there are no bindings
        but we wish to evaluate this derived predicate.  However, despite the fact
        that it has no bindings, we want to continue to (openly) solve predicates
        in a depth-first fashion until we hit an EDB query. 

        """
        evaluateTerm = Uniterm(BFP_NS.evaluate,
                               [label,
                                Literal(bodyLen)],
                               newNss=self.namespaces)
        return self.makeAdornedRule(
                And([self.makeDerivedQueryPredicate(rule.formula.head),
                     evaluateTerm]),rule.formula.head)

    def rule2(self,rule,label,bodyLen):
        """
        When a query is matched, collect answers (begin to evaluate)
        
        b^{k}
        
        evaluate(ruleNo,0,X) :- query_p(x0,...,xn) 
                                OpenQuery(query_p)
        
        If there are no bindings posed with the query, then OpenQuery(query_p) 
        is used instead of query_p(x0,...,xn), indicating that there are no bindings
        but we wish to evaluate this derived predicate.  However, despite the fact
        that it has no bindings, we want to continue to (openly) solve predicates
        in a depth-first fashion until we hit an EDB query.
        
        """
        evaluateTerm = Uniterm(BFP_NS.evaluate,[label,Literal(0)],
                               newNss=self.namespaces)        
        return self.makeAdornedRule(self.makeDerivedQueryPredicate(rule.formula.head),evaluateTerm)


class BFPQueryTerm(Uniterm):
    def __init__(self,uterm,adornment=None,naf = False,builtin = None):
        self.adornment=adornment
        self.nsMgr=uterm.nsMgr if hasattr(uterm,'nsMgr') else None
        if builtin:
            newArgs=[builtin.argument,builtin.result]
            op = builtin.uri
            self.builtin = builtin
        else:
            newArgs=copy.deepcopy(uterm.arg)
            op = uterm.op
            self.builtin = None
        super(BFPQueryTerm, self).__init__(op,newArgs,naf=naf)

    def clone(self):
        return BFPQueryTerm(self,self.adornment,self.naf)

    def _recalculateHash(self):
        self._hash=hash(reduce(lambda x,y:str(x)+str(y),
            len(self.arg)==2 and self.toRDFTuple() or [self.op]+self.arg))        

    def __hash__(self):
        if self.adornment is None:
            return self._hash
        else:
            return self._hash ^ hash(reduce(lambda x,y:x+y,self.adornment))
            
    def finalize(self):
        if self.adornment:
            if self.hasBindings():
                if len(self.adornment) == 1:
                    #adorned predicate occurrence with one out of two arguments bound
                    #convert: It becomes a unary predicate
                    #(an rdf:type assertion)
                    self.arg[-1] = URIRef(GetOp(self)+'_query_'+first(self.adornment))
                    self.arg[0] = first(self.getDistinguishedVariables())
                    self.op = RDF.type
                elif ''.join(self.adornment) == 'bb':
                    #Two bound args
                    self.setOperator(URIRef(self.op+'_query_bb'))
                else:
                    #remove unbound argument, and reduce arity
                    singleArg = first(self.getDistinguishedVariables())
                    self.arg[-1] = URIRef(GetOp(self)+'_query_b')
                    self.arg[0]  = singleArg 
                    self.op = RDF.type
                    
            else:
                currentOp = GetOp(self)
                self.op = RDF.type
                self.arg = [currentOp,BFP_RULE.OpenQuery]
        else:
            if GetOp(self) != HIGHER_ORDER_QUERY:
                self.setOperator(URIRef(GetOp(self)+'_query'))
        self._recalculateHash()
        
    def hasBindings(self):
        if self.adornment:
            for idx,term in enumerate(GetArgs(self)):
                if self.adornment[idx]=='b':
                    return True
            return False

    def getDistinguishedVariables(self,varsOnly=False):
        if self.op == RDF.type:
            for idx,term in enumerate(GetArgs(self)):
                if self.adornment[idx] in ['b','fb','bf']:
                    if not varsOnly or isinstance(term,Variable):
                        yield term
        else:
            for idx,term in enumerate(self.arg):
                if self.adornment[idx]=='b':
                    if not varsOnly or isinstance(term,Variable):
                        yield term

    def getBindings(self,uniterm):
        rt={}
        for idx,term in enumerate(self.arg):
            goalArg=self.arg[idx]
            candidateArg=uniterm.arg[idx]
            if self.adornment is None or (self.adornment[idx]=='b' and isinstance(candidateArg,Variable)):
                #binding
                rt[candidateArg]=goalArg
        return rt

    def __repr__(self):
        if self.builtin:
            return repr(self.builtin)
        else:
            pred = self.normalizeTerm(self.op)
            if self.op == RDF.type:
                adornSuffix = ''# if self.adornment is None else '_'+self.adornment[0]
            else:
                adornSuffix= ''# if self.adornment is None else '_'+''.join(self.adornment)
            if self.op == RDF.type:
                return "%s%s(%s)"%(self.normalizeTerm(self.arg[-1]),
                                   adornSuffix,
                                   self.normalizeTerm(self.arg[0]))
            else:
                return "%s%s(%s)"%(pred,
                                     adornSuffix,
                                     ' '.join([self.normalizeTerm(i)
                                                for i in self.arg]))


class BackwardFixpointProcedureTests(unittest.TestCase):
	def setUp(self):
		pass


if __name__ == '__main__':
	unittest.main()


from pprint import pprint, pformat
from sets import Set
from FuXi.Rete import *
from FuXi.Syntax.InfixOWL import *
from FuXi.Rete.AlphaNode import SUBJECT,PREDICATE,OBJECT,VARIABLE
from FuXi.Rete.BetaNode import LEFT_MEMORY,RIGHT_MEMORY
from FuXi.Rete.RuleStore import N3RuleStore, SetupRuleStore
from FuXi.Rete.Util import renderNetwork,generateTokenSet
from FuXi.Horn.PositiveConditions import Uniterm, BuildUnitermFromTuple
from FuXi.LP.BackwardFixpointProcedure import BackwardFixpointProcedure
from FuXi.LP import IdentifyHybridPredicates
from FuXi.SPARQL.BackwardChainingStore import * 
from FuXi.DLP.ConditionalAxioms import AdditionalRules
from FuXi.Horn.HornRules import HornFromN3
from FuXi.DLP import MapDLPtoNetwork, non_DHL_OWL_Semantics
from FuXi.Rete.Magic import *
from FuXi.Rete.SidewaysInformationPassing import *
from FuXi.Rete.TopDown import PrepareSipCollection, SipStrategy
from FuXi.SPARQL import RDFTuplesToSPARQL, EDBQuery
from FuXi.Rete.Proof import ProofBuilder, PML, GMP_NS
from rdflib.Namespace import Namespace
from rdflib import plugin,RDF,RDFS,URIRef,URIRef
from rdflib.OWL import FunctionalProperty
from rdflib.store import Store
from cStringIO import StringIO
from rdflib.Graph import Graph,ReadOnlyGraphAggregate,ConjunctiveGraph
from rdflib.syntax.NamespaceManager import NamespaceManager
from glob import glob
from rdflib.sparql.bison import Parse
import unittest, os, time, itertools

RDFLIB_CONNECTION=''
RDFLIB_STORE='IOMemory'

CWM_NS    = Namespace("http://cwmTest/")
DC_NS     = Namespace("http://purl.org/dc/elements/1.1/")
STRING_NS = Namespace("http://www.w3.org/2000/10/swap/string#")
MATH_NS   = Namespace("http://www.w3.org/2000/10/swap/math#")
FOAF_NS   = Namespace("http://xmlns.com/foaf/0.1/") 
OWL_NS    = Namespace("http://www.w3.org/2002/07/owl#")
TEST_NS   = Namespace("http://metacognition.info/FuXi/DL-SHIOF-test.n3#")
LOG       = Namespace("http://www.w3.org/2000/10/swap/log#")
RDF_TEST  = Namespace('http://www.w3.org/2000/10/rdf-tests/rdfcore/testSchema#')
OWL_TEST  = Namespace('http://www.w3.org/2002/03owlt/testOntology#')
LIST      = Namespace('http://www.w3.org/2000/10/swap/list#')

queryNsMapping={'test':'http://metacognition.info/FuXi/test#',
                'rdf':'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'foaf':'http://xmlns.com/foaf/0.1/',
                'dc':'http://purl.org/dc/elements/1.1/',
                'rss':'http://purl.org/rss/1.0/',
                'rdfs':'http://www.w3.org/2000/01/rdf-schema#',
                'rdf':'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'owl':OWL_NS,
                'rdfs':RDF.RDFNS,
}

nsMap = {
  u'rdfs' :RDFS.RDFSNS,
  u'rdf'  :RDF.RDFNS,
  u'rete' :RETE_NS,
  u'owl'  :OWL_NS,
  u''     :TEST_NS,
  u'otest':OWL_TEST,
  u'rtest':RDF_TEST,
  u'foaf' :URIRef("http://xmlns.com/foaf/0.1/"),
  u'math' :URIRef("http://www.w3.org/2000/10/swap/math#"),
}

MANIFEST_QUERY = \
"""
SELECT ?status ?premise ?conclusion ?feature ?descr
WHERE {
  [ 
    a otest:PositiveEntailmentTest;
    otest:feature ?feature;
    rtest:description ?descr;
    rtest:status ?status;
    rtest:premiseDocument ?premise;
    rtest:conclusionDocument ?conclusion 
  ]
}"""
PARSED_MANIFEST_QUERY = Parse(MANIFEST_QUERY)

Features2Skip = [
    URIRef('http://www.w3.org/2002/07/owl#sameClassAs'),
]

NonNaiveSkip = [
    'OWL/oneOf/Manifest002.rdf', #see Issue 25
    'OWL/unionOf/Manifest002.rdf',                   # support for disjunctive horn logic 
]

MagicTest2Skip = [
    'OWL/oneOf/Manifest002.rdf',        #requires second order predicate derivation
    'OWL/oneOf/Manifest003.rdf',        #requires second order predicate derivation
    'OWL/disjointWith/Manifest001.rdf'  #requires second order predicate derivation
]


BFPTests2SKip = [
    'OWL/FunctionalProperty/Manifest002.rdf'       , #Haven't reconciled *all* 2nd order predicate queries
    'OWL/InverseFunctionalProperty/Manifest002.rdf', #  "         "        "    "
    # 'OWL/oneOf/Manifest002.rdf',                     #  "         "        "    "
    'OWL/oneOf/Manifest003.rdf',                     #  "         "        "    "
]

TopDownTests2Skip = [
    'OWL/FunctionalProperty/Manifest002.rdf', #requires second order predicate derivation 
    'OWL/FunctionalProperty/Manifest004.rdf',
    'OWL/InverseFunctionalProperty/Manifest002.rdf', 
    'OWL/InverseFunctionalProperty/Manifest004.rdf',
    'OWL/oneOf/Manifest003.rdf', #Requires quantification over predicate symbol (2nd order)    
    'OWL/AllDifferent/Manifest001.rdf' #Not sure why
]

Tests2Skip = [
      
    'OWL/InverseFunctionalProperty/Manifest001.rdf', #owl:sameIndividualAs deprecated
    'OWL/FunctionalProperty/Manifest001.rdf', #owl:sameIndividualAs deprecated
    'OWL/Nothing/Manifest002.rdf',# owl:sameClassAs deprecated
]

patterns2Skip = [
    'OWL/cardinality',
    'OWL/samePropertyAs'
]

def tripleToTriplePattern(graph,triple):
    return "%s %s %s"%tuple([renderTerm(graph,term) 
                                for term in triple])

def renderTerm(graph,term):
    if term == RDF.type:
        return ' a '
    else:
        try:
            return isinstance(term,BNode) and term.n3() or graph.qname(term)
        except:
            return term.n3()

class OwlTestSuite(unittest.TestCase):
    def setUp(self):
        rule_store, rule_graph, self.network = SetupRuleStore(makeNetwork=True)
        self.network.nsMap = nsBinds
        
    def tearDown(self):
        pass
    
    def calculateEntailments(self, factGraph):
        start = time.time()  
        self.network.feedFactsToAdd(generateTokenSet(factGraph))                    
        sTime = time.time() - start
        if sTime > 1:
            sTimeStr = "%s seconds"%sTime
        else:
            sTime = sTime * 1000
            sTimeStr = "%s milli seconds"%sTime
        print "Time to calculate closure on working memory: ",sTimeStr
        print self.network
        
        tNodeOrder = [tNode 
                        for tNode in self.network.terminalNodes 
                            if self.network.instanciations.get(tNode,0)]
        tNodeOrder.sort(key=lambda x:self.network.instanciations[x],reverse=True)
        for termNode in tNodeOrder:
            print termNode
            print "\t", termNode.clause
            print "\t\t%s instanciations"%self.network.instanciations[termNode]
    #                    for c in AllClasses(factGraph):
    #                        print CastClass(c,factGraph)
        print "=============="
        self.network.inferredFacts.namespace_manager = factGraph.namespace_manager
        return sTimeStr

    def MagicOWLProof(self,goals,rules,factGraph,conclusionFile):
        progLen = len(rules)
        magicRuleNo = 0
        dPreds = []
        for rule in AdditionalRules(factGraph):
            rules.append(rule)            
        if not GROUND_QUERY and REASONING_STRATEGY != 'gms':
            goalDict = dict([((Variable('SUBJECT'),goalP,goalO),goalS) 
                        for goalS,goalP,goalO in goals])
            goals = goalDict.keys()
        assert goals

        if REASONING_STRATEGY == 'gms':
            for rule in MagicSetTransformation(factGraph,
                                               rules,
                                               goals,
                                               dPreds):
                magicRuleNo+=1
                self.network.buildNetworkFromClause(rule)    
                self.network.rules.add(rule)
                if DEBUG:
                    print "\t", rule
            print "rate of reduction in the size of the program: ", (100-(float(magicRuleNo)/float(progLen))*100)
        
        if REASONING_STRATEGY in ['bfp','sld']:# and not GROUND_QUERY:
            reasoningAlg = TOP_DOWN_METHOD if REASONING_STRATEGY == 'sld' \
                           else BFP_METHOD
            topDownStore=TopDownSPARQLEntailingStore(
                            factGraph.store,
                            factGraph,
                            idb=rules,
                            DEBUG=DEBUG,
                            nsBindings=nsMap,
                            decisionProcedure = reasoningAlg,
                            identifyHybridPredicates = REASONING_STRATEGY == 'bfp')
            targetGraph = Graph(topDownStore)
            for pref,nsUri in nsMap.items():
                targetGraph.bind(pref,nsUri)
            start = time.time()                      
                        
            for goal in goals:
                queryLiteral = EDBQuery([BuildUnitermFromTuple(goal)],
                                        factGraph,
                                        None if GROUND_QUERY else [goal[0]])
                query = queryLiteral.asSPARQL()
                print "Goal to solve ", query
                rt=targetGraph.query(query,initNs=nsMap)
                if GROUND_QUERY:
                    self.failUnless(rt.askAnswer[0],"Failed top-down problem")
                else:                    
                    self.failUnless((goalDict[goal]) in rt, 
                                    "Failed top-down problem")
            sTime = time.time() - start
            if sTime > 1:
                sTimeStr = "%s seconds"%sTime
            else:
                sTime = sTime * 1000
                sTimeStr = "%s milli seconds"%sTime
            return sTimeStr
        elif REASONING_STRATEGY == 'gms':
            for goal in goals:
                adornedGoalSeed = AdornLiteral(goal).makeMagicPred()
                goal=adornedGoalSeed.toRDFTuple()
                if DEBUG:
                    print "Magic seed fact ", adornedGoalSeed
                factGraph.add(goal)
            timing=self.calculateEntailments(factGraph)
            for goal in goals:
                # self.failUnless(goal in self.network.inferredFacts or goal in factGraph,
                #                 "Failed GMS query")
                if goal not in self.network.inferredFacts and goal not in factGraph:
                    print "missing triple %s"%(pformat(goal))
                    from FuXi.Rete.Util import renderNetwork 
                    pprint(list(factGraph.adornedProgram))
                    # dot=renderNetwork(self.network,self.network.nsMap).write_jpeg('test-fail.jpeg')
                    self.network.reportConflictSet(True)
                    raise #Exception ("Failed test: "+feature)
                else:
                    print "=== Passed! ==="
            return timing
    def testOwl(self): 
        testData = {}       
        for manifest in glob('OWL/*/Manifest*.rdf'):
            if manifest in Tests2Skip:
                continue
            if (REASONING_STRATEGY is not None and manifest in NonNaiveSkip) or\
               (REASONING_STRATEGY == 'sld' and manifest in TopDownTests2Skip) or \
               (REASONING_STRATEGY == 'bfp' and manifest in BFPTests2SKip) or\
               (REASONING_STRATEGY == 'gms' and manifest in MagicTest2Skip):
                continue
            
            skip = False
            for pattern2Skip in patterns2Skip:
                if manifest.find(pattern2Skip) > -1:
                    skip = True
                    break
            if skip:
                continue            
            manifestStore = plugin.get(RDFLIB_STORE,Store)()
            manifestGraph = Graph(manifestStore)
            manifestGraph.parse(open(manifest))
            rt = manifestGraph.query(
                                      MANIFEST_QUERY,
                                      initNs=nsMap,
                                      DEBUG = False)
            #print list(manifestGraph.namespace_manager.namespaces())
            for status,premise,conclusion, feature, description in rt:
                if feature in Features2Skip:
                    continue
                premise = manifestGraph.namespace_manager.compute_qname(premise)[-1]
                conclusion = manifestGraph.namespace_manager.compute_qname(conclusion)[-1]
                premiseFile    = '/'.join(manifest.split('/')[:2]+[premise])
                conclusionFile = '/'.join(manifest.split('/')[:2]+[conclusion])
                print premiseFile
                print conclusionFile
                if status == 'APPROVED':
                    if SINGLE_TEST and premiseFile != SINGLE_TEST:
                        continue
                    assert os.path.exists('.'.join([premiseFile,'rdf'])) 
                    assert os.path.exists('.'.join([conclusionFile,'rdf']))
                    print "<%s> :- <%s>"%('.'.join([conclusionFile,'rdf']),
                                          '.'.join([premiseFile,'rdf']))
                    store = plugin.get(RDFLIB_STORE,Store)()
                    store.open(RDFLIB_CONNECTION)
                    factGraph = Graph(store)
                    factGraph.parse(open('.'.join([premiseFile,'rdf'])))
                    nsMap.update(dict([(k,v)
                                 for k,v in factGraph.namespaces()]))                    
                    if DEBUG:
                        print "## Source Graph ##\n", factGraph.serialize(format='n3')
                    Individual.factoryGraph=factGraph
                    
                    for c in AllClasses(factGraph):
                        if not isinstance(c.identifier,BNode):
                            print c.__repr__(True)       
                            
                    if feature in TopDownTests2Skip:
                        continue
                    print premiseFile,feature,description
                    program=list(HornFromN3(StringIO(non_DHL_OWL_Semantics)))
                    program.extend(self.network.setupDescriptionLogicProgramming(
                                                                 factGraph,
                                                                 addPDSemantics=False,
                                                                 constructNetwork=False))                        
                    print "Original program"
                    pprint(program)
                    timings=[]  

                    if REASONING_STRATEGY is None:
                        sTimeStr=self.calculateEntailments(factGraph)
                        expectedFacts = Graph(store)                    
                        for triple in expectedFacts.parse('.'.join([conclusionFile,'rdf'])):
                            closureGraph = ReadOnlyGraphAggregate([self.network.inferredFacts,factGraph])
                            if triple not in self.network.inferredFacts and triple not in factGraph:
                                print "missing triple %s"%(pformat(triple))
                                print manifest
                                print "feature: ", feature
                                print description
                                pprint(list(self.network.inferredFacts))  
                                raise Exception ("Failed test: "+feature)
                            else:
                                print "=== Passed! ==="
                                #pprint(list(self.network.inferredFacts))
                        print "\n"
                        testData[manifest] = sTimeStr
                        store.rollback()
                        
                        self.setUp()                        
                        # self.network.reset()
                        # self.network._resetinstanciationStats()
                    else:
                        try:   
                            goals=[]
                            for triple in Graph(store).parse('.'.join([conclusionFile,'rdf'])):
                                if triple not in factGraph:
                                    goals.append(triple)
                            testData[manifest] = self.MagicOWLProof(goals,
                                                              program,
                                                              factGraph,
                                                              conclusionFile)
                                                              
                            self.setUp()
                            # self.network._resetinstanciationStats()
                            # self.network.reset()
                            # self.network.clear()                                
                        except:
    #                            print "missing triple %s"%(pformat(goal))
                            print manifest, premiseFile
                            print "feature: ", feature
                            print description
                            from FuXi.Rete.Util import renderNetwork 
                            pprint([BuildUnitermFromTuple(t) for t in self.network.inferredFacts])
    #                            dot=renderNetwork(self.network,self.network.nsMap).write_jpeg('test-fail.jpeg')
                            raise #Exception ("Failed test: "+feature)
                        
        pprint(testData)
        
def runTests(options):
    global REASONING_STRATEGY, GROUND_QUERY, SINGLE_TEST, DEBUG 
    SINGLE_TEST        = options.singleTest   
    DEBUG              = options.debug
    GROUND_QUERY       = options.groundQuery
    REASONING_STRATEGY = options.strategy

    suite = unittest.makeSuite(OwlTestSuite)
    if options.profile:
        #from profile import Profile
        from hotshot import Profile, stats
        p = Profile('fuxi.profile')
        #p = Profile()
        for i in range(options.runs):
            p.runcall(unittest.TextTestRunner(verbosity=5).run,suite)
        p.close()    
        s = stats.load('fuxi.profile')
#        s=p.create_stats()
        s.strip_dirs()
        s.sort_stats('time','cumulative','pcalls')
        s.print_stats(.1)
        s.print_callers(.05)
        s.print_callees(.05)
    else:
        for i in range(options.runs):        
            unittest.TextTestRunner(verbosity=5).run(suite)
                
if __name__ == '__main__':
    from optparse import OptionParser
    op = OptionParser('usage: %prog [options]')
    op.add_option('--profile', 
                  action='store_true',
                  default=False,
      help = 'Whether or not to run a profile')    
    op.add_option('--singleTest', 
                  default='',
      help = 'The identifier for the test to run')        
    op.add_option('--debug','-v', 
                  action='store_true',
                  default=False,
      help = 'Run the test in verbose mode')            
    op.add_option('--runs', 
                  type='int',
                  default=1,
      help = 'The number of times to run the test to accumulate data for profiling')            
    op.add_option('--groundQuery', 
                action='store_true',
                default=False,                  
      help = 'For top-down strategies, whether to solve ground triple patterns or not')
      
    op.add_option('--strategy', 
                  default='gms',
                  choices = ['gms','sld','bfp'],
      help = 'Which reasoning strategy to use in solving the OWL test cases '+
      'Use RETE-UL over re-written rules (gms), resolution-based top-down unification (sld),'+
      'or use RETE-UL as a backward fixpoint procedure (bfp) to simulate a top-down strategy via metainterpreter')
    (options, facts) = op.parse_args()
    
    runTests(options)


#!/usr/bin/env python
import unittest
from pprint import pprint
from cStringIO import StringIO
from rdflib import Namespace, ConjunctiveGraph, Literal, Variable, URIRef
from rdflib.Graph import QuotedGraph
from FuXi.Rete.BuiltinPredicates import FILTERS, STRING_NS
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.Rete.Util import generateTokenSet
from FuXi.Horn.HornRules import HornFromN3, NetworkFromN3
from FuXi.Rete.BuiltinPredicates import FILTERS, STRING_NS

TEST_NS = Namespace("http://example.org/test#")
LOG = Namespace("http://www.w3.org/2000/10/swap/log#")

def StringStartsWith(subject, object_):
    for term in (subject, object_):
        assert isinstance(term, (Variable, Literal, URIRef)), \
            "str:startsWith terms must be Variables, Literals, or URIRefs!"
    def startsWithF(subject_value, object_value):
        return subject_value.startswith(object_value)
    return startsWithF

def extractBaseFacts(cg):
    """
    Takes a conjunctive graph and returns
    a generator of RDF facts (excluding N3 facts
    involving log:implies and triples in QuotedGraphs)
    """
    for ctx in cg.contexts():
        if not isinstance(ctx,QuotedGraph):
            for s,p,o in ctx:
                if p != LOG.implies:
                    yield s,p,o

def build_network(rules):
    if isinstance(rules, basestring):
        rules = StringIO(rules)
    graph = ConjunctiveGraph()
    graph.load(rules, publicID='test', format='n3')
    network = NetworkFromN3(graph,
                            additionalBuiltins={STRING_NS.startsWith:StringStartsWith})
    network.feedFactsToAdd(generateTokenSet(extractBaseFacts(graph)))
    return network

def build_network2(rules):
    graph = ConjunctiveGraph()
    graph.load(StringIO(rules), publicID='test', format='n3')    
    rule_store, rule_graph=SetupRuleStore(
                      StringIO(rules),
                      additionalBuiltins={STRING_NS.startsWith:StringStartsWith})
    from FuXi.Rete.Network import ReteNetwork
    network = ReteNetwork(rule_store)
    network.feedFactsToAdd(generateTokenSet(extractBaseFacts(graph)))
    return network

class LiteralStringStartsWith(unittest.TestCase):
    fact = (TEST_NS.test, TEST_NS.passes, Literal(1))
    rules = """
    @prefix test: <http://example.org/test#> .
    @prefix str: <http://www.w3.org/2000/10/swap/string#> .

    test:example test:value "example" .
    { test:example test:value ?value .
      ?value str:startsWith "ex" } => { test:test test:passes 1 } .
    """
    def setUp(self):
        self.network  = build_network(self.rules)
        self.network2 = build_network2(self.rules)

    def test_literal_variable_startswith_literal_should_match(self):
        self.failUnless(self.fact in self.network.inferredFacts)
    def test_literal_variable_startswith_literal_should_match2(self):        
        self.failUnless(self.fact in self.network2.inferredFacts)

class URIRefStringStartsWith(unittest.TestCase):
    fact = (TEST_NS.test, TEST_NS.passes, Literal(1))
    rules = """
    @prefix test: <http://example.org/test#> .
    @prefix str: <http://www.w3.org/2000/10/swap/string#> .

    test:example test:value test:example .
    { test:example test:value ?value .
      ?value str:startsWith "http://example.org/test#ex" } => { test:test test:passes 1 } .
    """
    def setUp(self):
        self.network  = build_network(self.rules)
        self.network2 = build_network2(self.rules)

    def test_uriref_variable_startswith_literal_should_match(self):
        self.failUnless(self.fact in self.network.inferredFacts)
    def test_uriref_variable_startswith_literal_should_match2(self):        
        self.failUnless(self.fact in self.network2.inferredFacts)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
testSeralizationOfEval.py

Created by Chimezie Ogbuji on 2010-08-15.
Copyright (c) 2010 __MyCompanyName__. All rights reserved.
"""

import sys, os, unittest
from BackwardFixpointProcedure import BFP_RULE, BFP_NS
from FuXi.Horn.PositiveConditions import Uniterm
from FuXi.Horn.HornRules import Rule, Clause
from rdflib import Literal, RDF, Variable

nsBindings = {
    u'bfp'  : BFP_NS, 
    u'rule' : BFP_RULE 
}

class testSeralizationOfEvalTests(unittest.TestCase):
    def testSerializingEvalPred(self):
        evaluateTerm = Uniterm(BFP_NS.evaluate,
                               [BFP_RULE[str(1)],
                                Literal(1)],
                               newNss=nsBindings)
        self.failUnless(repr(evaluateTerm),"bfp:evaluate(rule:1 1)")
        xVar = Variable('X')
        yVar = Variable('Y')
        bodyTerm = Uniterm(RDF.rest,
                           [xVar,
                            yVar],
                           newNss=nsBindings)
        rule = Rule(Clause(bodyTerm,evaluateTerm),declare=[xVar,yVar])
        self.assertEqual(repr(rule),"Forall ?X ?Y ( bfp:evaluate(rule:1 1) :- rdf:rest(?X ?Y) )")

if __name__ == '__main__':
	unittest.main()


#!/usr/local/bin/python
# -*- coding: utf-8 -*-
"""
This module implements a Description Horn Logic implementation as defined
by Grosof, B. et.al. ("Description Logic Programs: Combining Logic Programs with 
Description Logic" [1]) in section 4.4.  As such, it implements recursive mapping
functions "T", "Th" and "Tb" which result in "custom" (dynamic) rulesets, RIF Basic 
Logic Dialect: Horn rulesets [2], [3].  The rulesets are evaluated against an 
efficient RETE-UL network.

It is a Description Logic Programming [1] Implementation on top of RETE-UL:

"A DLP is directly defined as the LP-correspondent of a def-Horn
ruleset that results from applying the mapping T ."

The mapping is as follows:

== Core (Description Horn Logic) ==

== Class Equivalence ==

T(owl:equivalentClass(C,D)) -> { T(rdfs:subClassOf(C,D) 
                                 T(rdfs:subClassOf(D,C) }
                                 
== Domain and Range Axioms (Base Description Logic: "ALC") ==                                                                                                       

T(rdfs:range(P,D))  -> D(y) := P(x,y)
T(rdfs:domain(P,D)) -> D(x) := P(x,y)

== Property Axioms (Role constructors: "I") ==

T(rdfs:subPropertyOf(P,Q))     -> Q(x,y) :- P(x,y)
T(owl:equivalentProperty(P,Q)) -> { Q(x,y) :- P(x,y)
                                    P(x,y) :- Q(x,y) }
T(owl:inverseOf(P,Q))          -> { Q(x,y) :- P(y,x)
                                    P(y,x) :- Q(x,y) }
T(owl:TransitiveProperty(P))   -> P(x,z) :- P(x,y) ^ P(y,z)                                                                        

[1] http://www.cs.man.ac.uk/~horrocks/Publications/download/2003/p117-grosof.pdf
[2] http://www.w3.org/2005/rules/wg/wiki/Core/Positive_Conditions
[3] http://www.w3.org/2005/rules/wg/wiki/asn06

"""

from __future__ import generators
from sets import Set
from rdflib import BNode, RDF, Namespace, Variable, RDFS
from rdflib.util import first
from rdflib.Collection import Collection
from rdflib.store import Store,VALID_STORE, CORRUPTED_STORE, NO_STORE, UNKNOWN
from rdflib import Literal, URIRef
from pprint import pprint, pformat
import sys, copy, itertools
from rdflib.Graph import QuotedGraph, Graph
from rdflib.store.REGEXMatching import REGEXTerm, NATIVE_REGEX, PYTHON_REGEX
from FuXi.Horn.PositiveConditions import And, Or, Uniterm, Condition, Atomic,SetOperator,Exists
from FuXi.Horn import DATALOG_SAFETY_NONE,DATALOG_SAFETY_STRICT, DATALOG_SAFETY_LOOSE
from LPNormalForms import NormalizeDisjunctions
from FuXi.Horn.HornRules import Clause as OriginalClause, Rule
from cStringIO import StringIO

SKOLEMIZED_CLASS_NS=Namespace('http://code.google.com/p/python-dlp/wiki/SkolemTerm#')

non_DHL_OWL_Semantics=\
"""
@prefix log: <http://www.w3.org/2000/10/swap/log#>.
@prefix math: <http://www.w3.org/2000/10/swap/math#>.
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix : <http://eulersharp.sourceforge.net/2003/03swap/owl-rules#>.
@prefix list: <http://www.w3.org/2000/10/swap/list#>.
#Additional OWL-compliant semantics, mappable to Production Rules 

#Subsumption (purely for TBOX classification)
{?C rdfs:subClassOf ?SC. ?A rdfs:subClassOf ?C} => {?A rdfs:subClassOf ?SC}.
{?C owl:equivalentClass ?A} => {?C rdfs:subClassOf ?A. ?A rdfs:subClassOf ?C}.
{?C rdfs:subClassOf ?SC. ?SC rdfs:subClassOf ?C} => {?C owl:equivalentClass ?SC}.

{?C owl:disjointWith ?B. ?M a ?C. ?Y a ?B } => {?M owl:differentFrom ?Y}.

{?P owl:inverseOf ?Q. ?P a owl:InverseFunctionalProperty} => {?Q a owl:FunctionalProperty}.
{?P owl:inverseOf ?Q. ?P a owl:FunctionalProperty} => {?Q a owl:InverseFunctionalProperty}.

#For OWL/InverseFunctionalProperty/premises004
{?C owl:oneOf ?L. ?L rdf:first ?X; rdf:rest rdf:nil. ?P rdfs:domain ?C} => {?P a owl:InverseFunctionalProperty}.
#For OWL/InverseFunctionalProperty/premises004
{?C owl:oneOf ?L. ?L rdf:first ?X; rdf:rest rdf:nil. ?P rdfs:range ?C} => {?P a owl:FunctionalProperty}.

{?S owl:differentFrom ?O} => {?O owl:differentFrom ?S}.
{?S owl:complementOf ?O} => {?O owl:complementOf ?S}.
{?S owl:disjointWith ?O} => {?O owl:disjointWith ?S}.

"""

OWL_NS    = Namespace("http://www.w3.org/2002/07/owl#")

LOG = Namespace("http://www.w3.org/2000/10/swap/log#")
Any = None

LHS = 0
RHS = 1

def reduceAnd(left,right):
    if isinstance(left,And):
        left = reduce(reduceAnd,left)
    elif isinstance(right,And):
        right = reduce(reduceAnd,right)
    if isinstance(left,list) and not isinstance(right,list):
        return left+[right]
    elif isinstance(left,list) and isinstance(right,list):
        return left+right
    elif isinstance(left,list) and not isinstance(right,list):
        return left+[right]
    elif not isinstance(left,list) and isinstance(right,list):
        return [left]+right
    else:
        return [left,right]
    
def NormalizeClause(clause):
    def fetchFirst(gen):
        rt=first(gen)
        assert rt is not None
        return rt
    if hasattr(clause.head,'next'):# and not isinstance(clause.head,Condition):
        clause.head = fetchFirst(clause.head)
    if hasattr(clause.body,'next'):# and not isinstance(clause.body,Condition):
        clause.body = fetchFirst(clause.body)
#    assert isinstance(clause.head,(Atomic,And,Clause)),repr(clause.head)
#    assert isinstance(clause.body,Condition),repr(clause.body)
    if isinstance(clause.head,And):
        clause.head.formulae = reduce(reduceAnd,clause.head,[])
    if isinstance(clause.body,And):
        clause.body.formulae = reduce(reduceAnd,clause.body)
#    print "Normalized clause: ", clause
#    assert clause.body is not None and clause.head is not None,repr(clause)
    return clause

class UnsupportedNegation(Exception):
    def __init__(self,msg):
        super(UnsupportedNegation, self).__init__(msg)

class Clause(OriginalClause):
    """
    The RETE-UL algorithm supports conjunctions of facts in the head of a rule
    i.e.:   H1 ^ H2 ^ ... ^ H3 :- B1 ^  ^ Bm
    The Clause definition is overridden to permit this syntax (not allowed
    in definite LP or Horn rules)
    
    In addition, since we allow (in definite Horn) entailments beyond simple facts
    we ease restrictions on the form of the head to include Clauses
        """
    def __init__(self,body,head):
        self.body = body
        self.head = head
        if isinstance(head,Uniterm):
            from FuXi.Rete.Network import HashablePatternList
            try:
                antHash=HashablePatternList([term.toRDFTuple() 
                                    for term in body],skipBNodes=True)
                consHash=HashablePatternList([term.toRDFTuple() 
                                    for term in head],skipBNodes=True)      
                self._bodyHash = hash(antHash)
                self._headHash = hash(consHash)             
                self._hash     = hash((self._headHash,self._bodyHash))                                                      
            except:
                self._hash = None
        else:
            self._hash = None

    def __hash__(self):
        if self._hash is None:
            from FuXi.Rete.Network import HashablePatternList
            antHash=HashablePatternList([term.toRDFTuple() 
                                for term in self.body],skipBNodes=True)
            consHash=HashablePatternList([term.toRDFTuple() 
                                for term in self.head],skipBNodes=True)                                                                                            
            self._bodyHash = hash(antHash)
            self._headHash = hash(consHash)             
            self._hash     = hash((self._headHash,self._bodyHash))                                                      
        return self._hash
                
    def __repr__(self):
        return "%r :- %r"%(self.head,self.body)

    def n3(self):
        return u'{ %s } => { %s }'%(self.body.n3(),self.head.n3())    

def makeRule(clause,nsMap):
    from FuXi.Horn.HornRules import Rule
    vars=set()
    for child in clause.head:
        if isinstance(child,Or):
            #Disjunction in the head, skip this rule:
            #When a disjunction occurs on the r.h.s. of a subclass axiom it 
            #becomes a disjunction in the head of the corresponding rule, and 
            #this cannot be handled within the def-Horn framework.            
            return None
        assert isinstance(child,Uniterm),repr(child)
        vars.update([term for term in child.toRDFTuple() if isinstance(term,Variable)])
    negativeStratus=False
    for child in clause.body:
        if hasattr(child,'naf') and child.naf:
            negativeStratus=True
        elif not hasattr(child,'naf'):
            child.naf = False
        vars.update([term for term in child.toRDFTuple() if isinstance(term,Variable)])
    return Rule(clause,declare=vars,nsMapping=nsMap,negativeStratus=negativeStratus)

def DisjunctiveNormalForm(program,safety = DATALOG_SAFETY_NONE, network = None):
    from FuXi.Rete.SidewaysInformationPassing import GetArgs, iterCondition, GetOp
    for rule in program:
        tx_horn_clause = NormalizeClause(rule.formula)
        for tx_horn_clause in LloydToporTransformation(tx_horn_clause,True):    
            if safety in [DATALOG_SAFETY_LOOSE, DATALOG_SAFETY_STRICT]:
                rule = Rule(tx_horn_clause,nsMapping=network and network.nsMap or {})
                if not rule.isSafe():
                    if safety == DATALOG_SAFETY_LOOSE:
                        import warnings
                        warnings.warn("Ignoring unsafe rule (%s)"%rule,
                                      SyntaxWarning,
                                      3)
                        continue
                    elif safety == DATALOG_SAFETY_STRICT:
                        raise SyntaxError("Unsafe RIF Core rule: %s"%rule) 
            disj = [i for i in breadth_first(tx_horn_clause.body) if isinstance(i,Or)]
            import warnings
            if len(disj)>0:
                NormalizeDisjunctions(disj,tx_horn_clause,program,network)
            elif isinstance(tx_horn_clause.head,(And,Uniterm)):
        #                print "No Disjunction in the body"
                    for hc in ExtendN3Rules(network,NormalizeClause(tx_horn_clause)):
                        yield makeRule(hc,network and network.nsMap or {})
    
def MapDLPtoNetwork(network,
                    factGraph,
                    complementExpansions=[],
                    constructNetwork=False,
                    derivedPreds=[],
                    ignoreNegativeStratus=False,
                    safety = DATALOG_SAFETY_NONE):
    from FuXi.Rete.SidewaysInformationPassing import GetArgs, iterCondition, GetOp
    ruleset=set()
    negativeStratus=[]
    for horn_clause in T(factGraph,complementExpansions=complementExpansions,derivedPreds=derivedPreds):
#        print "## RIF BLD Horn Rules: Before LloydTopor: ##\n",horn_clause
#        print "## RIF BLD Horn Rules: After LloydTopor: ##"
        fullReduce=True
        for tx_horn_clause in LloydToporTransformation(horn_clause,fullReduce):
            tx_horn_clause = NormalizeClause(tx_horn_clause)
            disj = [i for i in breadth_first(tx_horn_clause.body) if isinstance(i,Or)]
            import warnings
            if len(disj)>0:
                NormalizeDisjunctions(disj,
                                      tx_horn_clause,
                                      ruleset,
                                      network,
                                      constructNetwork,
                                      negativeStratus,
                                      ignoreNegativeStratus)
            elif isinstance(tx_horn_clause.head,(And,Uniterm)):
    #                print "No Disjunction in the body"
                    for hc in ExtendN3Rules(network,NormalizeClause(tx_horn_clause),constructNetwork):
                        if safety in [DATALOG_SAFETY_LOOSE, DATALOG_SAFETY_STRICT]:
                            rule = Rule(hc,nsMapping=network.nsMap)
                            if not rule.isSafe():
                                if safety == DATALOG_SAFETY_LOOSE:
                                    import warnings
                                    warnings.warn("Ignoring unsafe rule (%s)"%rule,
                                                  SyntaxWarning,
                                                  3)
                                    continue
                                elif safety == DATALOG_SAFETY_STRICT:
                                    raise SyntaxError("Unsafe RIF Core rule: %s"%rule) 
                        _rule=makeRule(hc,network.nsMap)
                        if _rule.negativeStratus:
                            negativeStratus.append(_rule)                    
                        if _rule is not None and (not _rule.negativeStratus or not ignoreNegativeStratus):
                            ruleset.add(_rule)                    
            #Extract free variables anre add rule to ruleset
#        print "#######################"
#    print "########## Finished Building decision network from DLP ##########"
    #renderNetwork(network).write_graphviz('out.dot')
    if ignoreNegativeStratus:
        return ruleset,negativeStratus
    else:
        return iter(ruleset)

def IsaFactFormingConclusion(head):
    """
    'Relative to the def-Horn ruleset, the def-LP is thus sound; moreover, it is 
    complete for fact-form conclusions, i.e., for queries whose answers amount 
    to conjunctions of facts. However, the def-LP is a mildly weaker version of 
    the def-Horn ruleset, in the following sense. Every conclusion of the def-LP
    must have the form of a fact. By contrast, the entailments, i.e., conclusions, 
    of the def-Horn ruleset are not restricted to be facts.' - Scan depth-first
    looking for Clauses
    """
    if isinstance(head,And):
        for i in head:
            if not IsaFactFormingConclusion(i):
                return False
        return True
    elif isinstance(head,Or):
        return False
    elif isinstance(head,Atomic):
        return True
    elif isinstance(head,OriginalClause):
        return False
    else:
        print head
        raise

def traverseClause(condition):
    if isinstance(condition,SetOperator):
        for i in iter(condition):
            yield i
    elif isinstance(condition,Atomic):
        return 

def breadth_first(condition,children=traverseClause):
    """Traverse the nodes of a tree in breadth-first order.
    The first argument should be the tree root; children
    should be a function taking as argument a tree node and
    returning an iterator of the node's children.
    
    From http://ndirty.cute.fi/~karttu/matikka/Python/eppsteins_bf_traversal_231503.htm
    
    """
    yield condition
    last = condition
    for node in breadth_first(condition,children):
        for child in children(node):
            yield child
            last = child
        if last == node:
            return

def breadth_first_replace(condition,
                          children=traverseClause,
                          candidate=None,
                          replacement=None):
    """Traverse the nodes of a tree in breadth-first order.
    The first argument should be the tree root; children
    should be a function taking as argument a tree node and
    returning an iterator of the node's children.
    
    From http://ndirty.cute.fi/~karttu/matikka/Python/eppsteins_bf_traversal_231503.htm
    
    """
    yield condition
    last = condition
    for node in breadth_first_replace(condition,
                                      children,
                                      candidate,
                                      replacement):
        for child in children(node):
            yield child
            if candidate and child is candidate:
                #replace candidate with replacement
                i=node.formulae.index(child)
                node.formulae[i]=replacement
                return
            last = child
        if last == node:
            return

def ExtendN3Rules(network,horn_clause,constructNetwork=False):
    """
    Extends the network with the given Horn clause (rule)
    """
    from FuXi.Rete.RuleStore import Formula
    from FuXi.Rete.AlphaNode import AlphaNode
    rt=[]
    if constructNetwork:
        ruleStore = network.ruleStore
        lhs = BNode()
        rhs = BNode()
    assert isinstance(horn_clause.body,(And,Uniterm)),list(horn_clause.body)
    assert len(list(horn_clause.body))
#    print horn_clause
    if constructNetwork:
        for term in horn_clause.body:
            ruleStore.formulae.setdefault(lhs,Formula(lhs)).append(term.toRDFTuple())
    assert isinstance(horn_clause.head,(And,Uniterm)),repr(horn_clause.head)

    if IsaFactFormingConclusion(horn_clause.head):
        PrepareHornClauseForRETE(horn_clause)
        if constructNetwork:
            for term in horn_clause.head:
                assert not hasattr(term,'next')
                if isinstance(term,Or):
                    ruleStore.formulae.setdefault(rhs,Formula(rhs)).append(term)
                else:
                    ruleStore.formulae.setdefault(rhs,Formula(rhs)).append(term.toRDFTuple())
            ruleStore.rules.append((ruleStore.formulae[lhs],ruleStore.formulae[rhs]))
            network.buildNetwork(iter(ruleStore.formulae[lhs]),
                                 iter(ruleStore.formulae[rhs]),
                                 Rule(horn_clause))
            network.alphaNodes = [node for node in network.nodes.values() if isinstance(node,AlphaNode)]
        rt.append(horn_clause)
    else:
        for hC in LloydToporTransformation(horn_clause,fullReduction=True):
            rt.append(hC)
            #print "normalized clause: ", hC
            for i in ExtendN3Rules(network,hC,constructNetwork):
                rt.append(hC)
    return rt

def PrepareHornClauseForRETE(horn_clause):
    if isinstance(horn_clause,Rule):
        horn_clause=horn_clause.formula
    def extractVariables(term,existential=True):
        if isinstance(term,existential and BNode or Variable):
            yield term
        elif isinstance(term,Uniterm):
            for t in term.toRDFTuple():
                if isinstance(t,existential and BNode or Variable):
                    yield t
    from FuXi.Rete.SidewaysInformationPassing import iterCondition, GetArgs
            
    #first we identify body variables                        
    bodyVars = set(reduce(lambda x,y:x+y,
                          [ list(extractVariables(i,existential=False)) for i in iterCondition(horn_clause.body) ]))
    
    #then we identify head variables
    headVars = set(reduce(lambda x,y:x+y,
                          [ list(extractVariables(i,existential=False)) for i in iterCondition(horn_clause.head) ]))
    
    #then we identify those variables that should (or should not) be converted to skolem terms
    updateDict       = dict([(var,BNode()) for var in headVars if var not in bodyVars])
    
    if set(updateDict.keys()).intersection(GetArgs(horn_clause.head)):
        #There are skolem terms in the head
        newHead = copy.deepcopy(horn_clause.head)
        for uniTerm in iterCondition(newHead):
            newArg      = [ updateDict.get(i,i) for i in uniTerm.arg ]
            uniTerm.arg = newArg
        horn_clause.head = newHead
        
    skolemsInBody=[
                   list(itertools.ifilter(
                             lambda term:isinstance(term,
                                                    BNode),
                                 GetArgs(lit))) 
                                 for lit in iterCondition(horn_clause.body)]
    skolemsInBody = reduce(lambda x,y:x+y,skolemsInBody,
                           [])
    if skolemsInBody:
        newBody = copy.deepcopy(horn_clause.body)
        _e=Exists(formula=newBody,declare=set(skolemsInBody))        
        horn_clause.body=_e

    skolemsInHead=[
                   list(itertools.ifilter(
                             lambda term:isinstance(term,
                                                    BNode),
                                 GetArgs(lit))) 
                                 for lit in iterCondition(horn_clause.head)]
    skolemsInHead = reduce(lambda x,y:x+y,skolemsInHead,
                           [])
    if skolemsInHead:
        newHead = copy.deepcopy(horn_clause.head)
        _e=Exists(formula=newHead,declare=set(skolemsInHead))        
        horn_clause.head=_e

def generatorFlattener(gen):
    assert hasattr(gen,'next')
    i = list(gen)
    i = len(i)>1 and [hasattr(i2,'next') and generatorFlattener(i2) or i2 for i2 in i] or i[0]
    if hasattr(i,'next'):
        i=listOrThingGenerator(i)
        #print i
        return i
    elif isinstance(i,SetOperator):
        i.formulae = [hasattr(i2,'next') and generatorFlattener(i2) or i2 for i2 in i.formulae]
        #print i
        return i
    else:
        return i

def SkolemizeExistentialClasses(term,check=True):
    if check:
        return isinstance(term,BNode) and SKOLEMIZED_CLASS_NS[term] or term
    return SKOLEMIZED_CLASS_NS[term]

def NormalizeBooleanClassOperand(term,owlGraph):
    return ((isinstance(term,BNode) and IsaBooleanClassDescription(term,owlGraph)) or \
             IsaRestriction(term,owlGraph))\
          and SkolemizeExistentialClasses(term) or term    

def IsaBooleanClassDescription(term,owlGraph):
    for s,p,o in owlGraph.triples_choices((term,[OWL_NS.unionOf,
                                                OWL_NS.intersectionOf],None)):
        return True

def IsaRestriction(term,owlGraph):
    return (term,RDF.type,OWL_NS.Restriction) in owlGraph

def iterCondition(condition):
    return isinstance(condition,SetOperator) and condition or iter([condition])

def Tc(owlGraph,negatedFormula):
    """
    Handles the conversion of negated DL concepts into a general logic programming
    condition for the body of a rule that fires when the body conjunct
    is in the minimal model
    """
    if (negatedFormula,OWL_NS.hasValue,None) in owlGraph:
        #not ( R value i )
        bodyUniTerm = Uniterm(RDF.type,
                              [Variable("X"),
                               NormalizeBooleanClassOperand(negatedFormula,owlGraph)],
                              newNss=owlGraph.namespaces())
        
        condition = NormalizeClause(Clause(Tb(owlGraph,negatedFormula),
                                           bodyUniTerm)).body
        assert isinstance(condition,Uniterm)
        condition.naf = True
        return condition
    elif (negatedFormula,OWL_NS.someValuesFrom,None) in owlGraph:
        #not ( R some C )
        binaryRel,unaryRel = Tb(owlGraph,negatedFormula)
        negatedBinaryRel = copy.deepcopy(binaryRel)
        negatedBinaryRel.naf = True
        negatedUnaryRel  = copy.deepcopy(unaryRel)
        negatedUnaryRel.naf = True
        return Or([negatedBinaryRel,And([binaryRel,negatedUnaryRel])])
    elif isinstance(negatedFormula,URIRef):
        return Uniterm(RDF.type,
                       [Variable("X"),
                        NormalizeBooleanClassOperand(negatedFormula,owlGraph)],
                       newNss=owlGraph.namespaces(),
                       naf=True)
    else:
        raise UnsupportedNegation("Unsupported negated concept: %s"%negatedFormula)
    
class MalformedDLPFormulaError(NotImplementedError):
    def __init__(self,message):
        self.message = message
    
def handleConjunct(conjunction,owlGraph,o,conjunctVar=Variable('X')):
    for bodyTerm in Collection(owlGraph,o):
        negatedFormula = False
        addToConjunct=None
        for negatedFormula in owlGraph.objects(subject=bodyTerm,
                                               predicate=OWL_NS.complementOf):
            addToConjunct = Tc(owlGraph,negatedFormula)
        if negatedFormula:
            #addToConjunct will be the term we need to add to the conjunct
            conjunction.append(addToConjunct)
        else:
            normalizedBodyTerm=NormalizeBooleanClassOperand(bodyTerm,owlGraph)
            bodyUniTerm = Uniterm(RDF.type,[conjunctVar,normalizedBodyTerm],
                                  newNss=owlGraph.namespaces())
            processedBodyTerm=Tb(owlGraph,bodyTerm,conjunctVar)
            classifyingClause = NormalizeClause(Clause(processedBodyTerm,bodyUniTerm))
            redundantClassifierClause = processedBodyTerm == bodyUniTerm
            if isinstance(normalizedBodyTerm,URIRef) and normalizedBodyTerm.find(SKOLEMIZED_CLASS_NS)==-1:
                conjunction.append(bodyUniTerm)
            elif (bodyTerm,OWL_NS.someValuesFrom,None) in owlGraph or\
                 (bodyTerm,OWL_NS.hasValue,None) in owlGraph:                    
                conjunction.extend(classifyingClause.body)
            elif (bodyTerm,OWL_NS.allValuesFrom,None) in owlGraph:
                raise MalformedDLPFormulaError("Universal restrictions can only be used as the second argument to rdfs:subClassOf (GCIs)")
            elif (bodyTerm,OWL_NS.unionOf,None) in owlGraph:
                conjunction.append(classifyingClause.body)
            elif (bodyTerm,OWL_NS.intersectionOf,None) in owlGraph:
                conjunction.append(bodyUniTerm)                    
                        
def T(owlGraph,complementExpansions=[],derivedPreds=[]):
    """
    #Subsumption (purely for TBOX classification)
    {?C rdfs:subClassOf ?SC. ?A rdfs:subClassOf ?C} => {?A rdfs:subClassOf ?SC}.
    {?C owl:equivalentClass ?A} => {?C rdfs:subClassOf ?A. ?A rdfs:subClassOf ?C}.
    {?C rdfs:subClassOf ?SC. ?SC rdfs:subClassOf ?C} => {?C owl:equivalentClass ?SC}.
    
    T(rdfs:subClassOf(C,D))       -> Th(D(y)) :- Tb(C(y))
    
    T(owl:equivalentClass(C,D)) -> { T(rdfs:subClassOf(C,D) 
                                     T(rdfs:subClassOf(D,C) }
    
    A generator over the Logic Programming rules which correspond
    to the DL  ( unary predicate logic ) subsumption axiom described via rdfs:subClassOf
    """
    for s,p,o in owlGraph.triples((None,OWL_NS.complementOf,None)):
        if isinstance(o,URIRef) and isinstance(s,URIRef):
            headLiteral = Uniterm(RDF.type,[Variable("X"),
                                            SkolemizeExistentialClasses(s)],
                                  newNss=owlGraph.namespaces())
            yield NormalizeClause(Clause(Tc(owlGraph,o),headLiteral))
    for c,p,d in owlGraph.triples((None,RDFS.subClassOf,None)):
        try:
            yield NormalizeClause(Clause(Tb(owlGraph,c),Th(owlGraph,d)))
        except UnsupportedNegation:
            import warnings
            warnings.warn("Unable to handle negation in DL axiom (%s), skipping"%c,#e.msg,
                          SyntaxWarning,
                          3)
        #assert isinstance(c,URIRef),"%s is a kind of %s"%(c,d)
    for c,p,d in owlGraph.triples((None,OWL_NS.equivalentClass,None)):
        if c not in derivedPreds:
            yield NormalizeClause(Clause(Tb(owlGraph,c),Th(owlGraph,d)))
        yield NormalizeClause(Clause(Tb(owlGraph,d),Th(owlGraph,c)))
    for s,p,o in owlGraph.triples((None,OWL_NS.intersectionOf,None)):
        try:
            if s not in complementExpansions:
                if s in derivedPreds:
                    import warnings
                    warnings.warn("Derived predicate (%s) is defined via a conjunction (consider using a complex GCI) "%owlGraph.qname(s),
                                  SyntaxWarning,
                                  3)
                elif isinstance(s,BNode):# and (None,None,s) not in owlGraph:# and \
                     #(s,RDFS.subClassOf,None) in owlGraph:
                        #complex GCI, pass over (handled) by Tb
                        continue
                conjunction = []
                handleConjunct(conjunction,owlGraph,o)
                body = And(conjunction)
                head = Uniterm(RDF.type,[Variable("X"),
                                         SkolemizeExistentialClasses(s)],
                                         newNss=owlGraph.namespaces())
    #            O1 ^ O2 ^ ... ^ On => S(?X)            
                yield Clause(body,head)
                if isinstance(s,URIRef):
    #                S(?X) => O1 ^ O2 ^ ... ^ On                
        #            special case, owl:intersectionOf is a neccessary and sufficient
        #            criteria and should thus work in *both* directions 
        #            This rule is not added for anonymous classes or derived predicates
                    if s not in derivedPreds:
                        yield Clause(head,body)
        except UnsupportedNegation:
            import warnings
            warnings.warn("Unable to handle negation in DL axiom (%s), skipping"%s,#e.msg,
                          SyntaxWarning,
                          3)
        
    for s,p,o in owlGraph.triples((None,OWL_NS.unionOf,None)):
        if isinstance(s,URIRef):
            #special case, owl:unionOf is a neccessary and sufficient
            #criteria and should thus work in *both* directions
            body = Or([Uniterm(RDF.type,[Variable("X"),
                                         NormalizeBooleanClassOperand(i,owlGraph)],
                                         newNss=owlGraph.namespaces()) \
                           for i in Collection(owlGraph,o)])
            head = Uniterm(RDF.type,[Variable("X"),s],newNss=owlGraph.namespaces())
            yield Clause(body,head)
    for s,p,o in owlGraph.triples((None,OWL_NS.inverseOf,None)):
        #    T(owl:inverseOf(P,Q))          -> { Q(x,y) :- P(y,x)
        #                                        P(y,x) :- Q(x,y) }
        newVar = Variable(BNode())
        body1 = Uniterm(s,[newVar,Variable("X")],newNss=owlGraph.namespaces())
        head1 = Uniterm(o,[Variable("X"),newVar],newNss=owlGraph.namespaces())
        yield Clause(body1,head1)
        newVar = Variable(BNode())
        body2 = Uniterm(o,[Variable("X"),newVar],newNss=owlGraph.namespaces())
        head2 = Uniterm(s,[newVar,Variable("X")],newNss=owlGraph.namespaces())
        yield Clause(body2,head2)
    for s,p,o in owlGraph.triples((None,RDF.type,OWL_NS.TransitiveProperty)):
        #T(owl:TransitiveProperty(P))   -> P(x,z) :- P(x,y) ^ P(y,z)
        y = Variable(BNode())
        z = Variable(BNode())
        x = Variable("X")
        body = And([Uniterm(s,[x,y],newNss=owlGraph.namespaces()),\
                    Uniterm(s,[y,z],newNss=owlGraph.namespaces())])
        head = Uniterm(s,[x,z],newNss=owlGraph.namespaces())
        yield Clause(body,head)
    for s,p,o in owlGraph.triples((None,RDFS.subPropertyOf,None)):
        # T(rdfs:subPropertyOf(P,Q))     -> Q(x,y) :- P(x,y)        
        x = Variable("X")
        y = Variable("Y")
        body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
        head = Uniterm(o,[x,y],newNss=owlGraph.namespaces())
        yield Clause(body,head)
    for s,p,o in owlGraph.triples((None,OWL_NS.equivalentProperty,None)):
        # T(owl:equivalentProperty(P,Q)) -> { Q(x,y) :- P(x,y)
        #                                     P(x,y) :- Q(x,y) }
        x = Variable("X")
        y = Variable("Y")
        body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
        head = Uniterm(o,[x,y],newNss=owlGraph.namespaces())
        yield Clause(body,head)
        yield Clause(head,body)

    #Contribution (Symmetric DL roles)
    for s,p,o in owlGraph.triples((None,RDF.type,OWL_NS.SymmetricProperty)):
        #T(owl:SymmetricProperty(P))   -> P(y,x) :- P(x,y)
        y = Variable("Y")
        x = Variable("X")
        body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
        head = Uniterm(s,[y,x],newNss=owlGraph.namespaces())
        yield Clause(body,head)
        
    for s,p,o in owlGraph.triples_choices((None,
                                           [RDFS.range,RDFS.domain],
                                           None)):
        if p == RDFS.range:
            #T(rdfs:range(P,D))  -> D(y) := P(x,y)        
            x = Variable("X")
            y = Variable(BNode())
            body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
            head = Uniterm(RDF.type,[y,o],newNss=owlGraph.namespaces())
            yield Clause(body,head)
        else: 
            #T(rdfs:domain(P,D)) -> D(x) := P(x,y)
            x = Variable("X")
            y = Variable(BNode())
            body = Uniterm(s,[x,y],newNss=owlGraph.namespaces())
            head = Uniterm(RDF.type,[x,o],newNss=owlGraph.namespaces())
            yield Clause(body,head)
            
def LloydToporTransformation(clause,fullReduction=True):
    """
    Tautological, common horn logic forms (useful for normalizing 
    conjunctive & disjunctive clauses)
    
    (H ^ H0) :- B                 -> { H  :- B
                                       H0 :- B }
    (H :- H0) :- B                -> H :- B ^ H0
    
    H :- (B v B0)                 -> { H :- B
                                       H :- B0 }
    """
    assert isinstance(clause,OriginalClause),repr(clause)
    assert isinstance(clause.body,Condition),repr(clause)
    if isinstance(clause.body,Or):
        for atom in clause.body.formulae:
            if hasattr(atom, 'next'):
                atom=first(atom)
            yield NormalizeClause(Clause(atom,clause.head))
    elif isinstance(clause.head,OriginalClause):
        yield NormalizeClause(Clause(And([clause.body,clause.head.body]),clause.head.head))
    elif fullReduction and isinstance(clause.head,And):
        for i in clause.head:
            for j in LloydToporTransformation(Clause(clause.body,i),
                                              fullReduction=fullReduction):
                if [i for i in breadth_first(j.head) if isinstance(i,And)]:
                    #Ands in the head need to be further flattened
                    yield NormalizeClause(j) 
                else:
                    yield j
    else:
        yield clause
    

def Th(owlGraph,_class,variable=Variable('X'),position=LHS):
    """
    DLP head (antecedent) knowledge assertional forms (ABox assertions, conjunction of
    ABox assertions, and universal role restriction assertions)
    """
    props = list(set(owlGraph.predicates(subject=_class)))
    if OWL_NS.allValuesFrom in props:
        #http://www.w3.org/TR/owl-semantics/#owl_allValuesFrom
        for s,p,o in owlGraph.triples((_class,OWL_NS.allValuesFrom,None)):
            prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
            newVar = Variable(BNode())
            body = Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces())
            for head in Th(owlGraph,o,variable=newVar):
                yield Clause(body,head)
    elif OWL_NS.hasValue in props:
        prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
        o =first(owlGraph.objects(subject=_class,predicate=OWL_NS.hasValue))
        yield Uniterm(prop,[variable,o],newNss=owlGraph.namespaces())
    elif OWL_NS.someValuesFrom in props:
        #http://www.w3.org/TR/owl-semantics/#someValuesFrom
        for s,p,o in owlGraph.triples((_class,OWL_NS.someValuesFrom,None)):
            prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
            newVar = BNode()
            yield And([Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces()),
                        generatorFlattener(Th(owlGraph,o,variable=newVar))])
    elif OWL_NS.intersectionOf in props:
        from FuXi.Syntax.InfixOWL import BooleanClass
        yield And([first(Th(owlGraph,h,variable)) for h in BooleanClass(_class)])
    else:
        #Simple class
        yield Uniterm(RDF.type,[variable,
                                isinstance(_class,BNode) and SkolemizeExistentialClasses(_class) or _class],
                                newNss=owlGraph.namespaces())
            
    
def Tb(owlGraph,_class,variable=Variable('X')):
    """
    DLP body (consequent knowledge assertional forms (ABox assertions, 
    conjunction / disjunction of ABox assertions, and exisential role restriction assertions)
    These are all common EL++ templates for KR
    """
    props = list(set(owlGraph.predicates(subject=_class)))
    if OWL_NS.intersectionOf in props and not isinstance(_class,URIRef):
        for s,p,o in owlGraph.triples((_class,OWL_NS.intersectionOf,None)):
            conj=[]
            handleConjunct(conj,owlGraph,o,variable)
            return And(conj)
    elif OWL_NS.unionOf in props and not isinstance(_class,URIRef):
        #http://www.w3.org/TR/owl-semantics/#owl_unionOf
        for s,p,o in owlGraph.triples((_class,OWL_NS.unionOf,None)):
            return Or([Tb(owlGraph,c,variable=variable) \
                           for c in Collection(owlGraph,o)])
    elif OWL_NS.someValuesFrom in props:
        #http://www.w3.org/TR/owl-semantics/#owl_someValuesFrom
        prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
        o =list(owlGraph.objects(subject=_class,predicate=OWL_NS.someValuesFrom))[0]
        newVar = Variable(BNode())
        body = Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces())
        head = Th(owlGraph,o,variable=newVar)
        return And([Uniterm(prop,[variable,newVar],newNss=owlGraph.namespaces()),
                    Tb(owlGraph,o,variable=newVar)])
    elif OWL_NS.hasValue in props:
        #http://www.w3.org/TR/owl-semantics/#owl_hasValue
        #Domain-specific rules for hasValue
        #Can be achieved via pD semantics        
        prop = list(owlGraph.objects(subject=_class,predicate=OWL_NS.onProperty))[0]
        o =first(owlGraph.objects(subject=_class,predicate=OWL_NS.hasValue))
        return Uniterm(prop,[variable,o],newNss=owlGraph.namespaces())
    elif OWL_NS.complementOf in props:
        return Tc(owlGraph,first(owlGraph.objects(_class,OWL_NS.complementOf)))
    else:
        #simple class
        #"Named" Uniterm
        _classTerm=SkolemizeExistentialClasses(_class)
        return Uniterm(RDF.type,[variable,_classTerm],newNss=owlGraph.namespaces())


#!/usr/local/bin/python
# -*- coding: utf-8 -*-
"""
"""
import copy, warnings
from rdflib import BNode, RDF, Namespace, Variable, RDFS
from FuXi.Horn.PositiveConditions import And, Or, Uniterm, PredicateExtentFactory, SetOperator,Exists

def HasNestedConjunction(conjunct):
    rt=False
    for item in conjunct:
        if isinstance(item,And):
            rt=True
            break
    return rt

def flattenHelper(condition):
    toDo = [item for item in condition if isinstance(item,And)]
    for i in toDo:
        condition.formulae.remove(i)
    for i in toDo:
        condition.formulae.extend(i)

def HasBreadthFirstNestedConj(condition):
    from FuXi.DLP import breadth_first
    return HasNestedConjunction(condition) or\
           [i for i in breadth_first(condition) 
                if isinstance(i,And) and HasNestedConjunction(i)]

def FlattenConjunctions(condition,isNested=False):
    from FuXi.DLP import breadth_first
    if isNested or HasNestedConjunction(condition):
        flattenHelper(condition)
    for nestedConj in [i for i in breadth_first(condition) 
                       if isinstance(i,And) and HasNestedConjunction(i)]:
        FlattenConjunctions(nestedConj, isNested=True)    

def ApplyDemorgans(clause):
    """
    >>> from FuXi.DLP import Clause
    >>> EX_NS = Namespace('http://example.com/')
    >>> ns = {'ex' : EX_NS}
    >>> pred1 = PredicateExtentFactory(EX_NS.somePredicate,newNss=ns)
    >>> pred2 = PredicateExtentFactory(EX_NS.somePredicate2,newNss=ns)
    >>> pred3 = PredicateExtentFactory(EX_NS.somePredicate3,binary=False,newNss=ns)
    >>> pred4 = PredicateExtentFactory(EX_NS.somePredicate4,binary=False,newNss=ns)
    >>> clause = Clause(And([pred1[(Variable('X'),Variable('Y'))],
    ...                      Or([pred2[(Variable('X'),EX_NS.individual1)],
    ...                          pred3[(Variable('Y'))]],naf=True)]),
    ...                 pred4[Variable('X')])
    >>> clause
    ex:somePredicate4(?X) :- And( ex:somePredicate(?X ?Y) not Or( ex:somePredicate2(?X ex:individual1) ex:somePredicate3(?Y) ) )
    >>> ApplyDemorgans(clause)
    >>> clause
    ex:somePredicate4(?X) :- And( ex:somePredicate(?X ?Y) And( not ex:somePredicate2(?X ex:individual1) not ex:somePredicate3(?Y) ) )
    >>> FlattenConjunctions(clause.body)
    >>> clause
    ex:somePredicate4(?X) :- And( ex:somePredicate(?X ?Y) not ex:somePredicate2(?X ex:individual1) not ex:somePredicate3(?Y) )
    """
    from FuXi.DLP import breadth_first, breadth_first_replace
    replacementMap = {}
    for negDisj in [i for i in breadth_first(clause.body) 
                        if isinstance(i,Or) and i.naf]:
        replacementList = []
        for innerTerm in negDisj:
            assert isinstance(i,Uniterm)
            innerTerm.naf = not innerTerm.naf
            replacementList.append(innerTerm)
        replacementMap[negDisj] = And(replacementList)
    for old,new in replacementMap.items():
        list(breadth_first_replace(clause.body,candidate=old,replacement=new))
        
def HandleNonDisjunctiveClauses(ruleset, network, constructNetwork, negativeStratus, ignoreNegativeStratus, clause):
    from FuXi.DLP import NormalizeClause, ExtendN3Rules, makeRule
    for hc in ExtendN3Rules(network, NormalizeClause(clause), constructNetwork):
        rule = makeRule(hc, network.nsMap)
        if rule.negativeStratus:
            negativeStratus.append(rule)
        if not rule.negativeStratus or not ignoreNegativeStratus:
            ruleset.add(rule)
            
def NormalizeDisjunctions(disj,
                          clause,
                          ruleset,
                          network,
                          constructNetwork,
                          negativeStratus,
                          ignoreNegativeStratus):
    """
    Removes disjunctions from logic programs (if possible)
    """
    from FuXi.DLP import breadth_first, breadth_first_replace
#    disj = [i for i in breadth_first(clause.body) if isinstance(i,Or)]
    while len(disj) > 1:
        ApplyDemorgans(clause)
        if HasBreadthFirstNestedConj(clause.body):
            FlattenConjunctions(clause.body)
        disj=[i for i in breadth_first(clause.body) if isinstance(i,Or)]
        assert len(disj)<2,"Unable to effectively reduce disjunctions"
    if len(disj) == 1:
        #There is one disjunction in the body, we can reduce from:
        #H :- B1 V B2  to H : - B1 and H :- B2 
        origDisj = disj[0]
        for item in origDisj:
            #First we want to replace the entire disjunct with an item within it
            list(breadth_first_replace(clause.body,candidate=origDisj,replacement=item))
            clause_clone = copy.deepcopy(clause)
            disj = [i for i in breadth_first(clause_clone.body) if isinstance(i,Or)]
            if len(disj) > 0:
                #If the formula has disjunctions of it's own, we handle them recursively
                NormalizeDisjunctions(disj,
                                      clause_clone,
                                      ruleset,
                                      network,
                                      constructNetwork,
                                      negativeStratus,
                                      ignoreNegativeStratus)
            else:               
                if HasBreadthFirstNestedConj(clause_clone.body):
                    FlattenConjunctions(clause_clone.body)
                #Otherwise handle normally
                HandleNonDisjunctiveClauses(ruleset, network, constructNetwork, negativeStratus, ignoreNegativeStratus, clause_clone)
            #restore the replaced term (for the subsequent iteration)
            list(breadth_first_replace(clause.body,candidate=item,replacement=origDisj))
    else:
        #The disjunction has been handled by normal form transformation, we just need to 
        #handle normally
        if HasBreadthFirstNestedConj(clause_clone.body):
            FlattenConjunctions(clause_clone.body)        
        HandleNonDisjunctiveClauses(ruleset, network, constructNetwork, negativeStratus, ignoreNegativeStratus, clause)
    
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()    


import itertools
from FuXi.Rete.Proof import *
from rdflib import RDFS, RDF, Variable
from rdflib.util import first
from rdflib.store import Store
from rdflib.store.REGEXMatching import NATIVE_REGEX
from rdflib.sparql.Algebra import *
from rdflib.sparql.graphPattern import BasicGraphPattern
from rdflib.sparql.bison.Query import Query
from rdflib.Graph import Graph
from FuXi.DLP import DisjunctiveNormalForm
from FuXi.Rete.Magic import *
from FuXi.Rete.TopDown import *
from FuXi.Rete.Network import ReteNetwork
from FuXi.Rete.SidewaysInformationPassing import *
from FuXi.LP.BackwardFixpointProcedure import BackwardFixpointProcedure
from FuXi.LP import IdentifyHybridPredicates
from FuXi.Horn.PositiveConditions import BuildUnitermFromTuple
from FuXi.Rete.Util import lazyGeneratorPeek

TOP_DOWN_METHOD = 0
BFP_METHOD      = 1

DEFAULT_BUILTIN_MAP = { LOG.equal:       "%s  = %s",
                        LOG.notEqualTo:  "%s != %s" }

class TopDownSPARQLEntailingStore(Store):
    """
    A Store which uses FuXi's magic set "sip strategies" and the in-memory SPARQL Algebra
    implementation as a store-agnostic, top-down decision procedure for 
    semanic web SPARQL (OWL2-RL/RIF/N3) entailment regimes.  Exposed
    as a rdflib / layercake-python API for SPARQL datasets with entailment regimes
    Queries are mediated over the SPARQL protocol using global schemas captured
    as SW theories which describe and distinguish their predicate symbols
    """
    context_aware = True
    formula_aware = True
    transaction_aware = True
    regex_matching = NATIVE_REGEX
    batch_unification = True
    def getDerivedPredicates(self,expr,prolog):
        if isinstance(expr,BasicGraphPattern):
            for s,p,o,func in expr.patterns:
                derivedPred = self.derivedPredicateFromTriple((s,p,o))
                if derivedPred is not None:
                    yield derivedPred
        elif isinstance(expr,NonSymmetricBinaryOperator):
            for term in self.getDerivedPredicates(expr.left,prolog):
                yield term
            for term in self.getDerivedPredicates(expr.right,prolog):
                yield term
        else:
            raise NotImplementedError(expr)
    
    def isaBaseQuery(self, queryString,queryObj=None):
        """
        If the given SPARQL query involves purely base predicates 
        it returns it (as a parsed string), otherwise it returns a SPARQL algebra
        instance for top-down evaluation using this store
        
        >>> graph=Graph()
        >>> topDownStore = TopDownSPARQLEntailingStore(graph.store,[RDFS.seeAlso],nsBindings={u'rdfs':RDFS.RDFSNS})
        >>> rt=topDownStore.isaBaseQuery("SELECT * { [] rdfs:seeAlso [] }")
        >>> isinstance(rt,(BasicGraphPattern,AlgebraExpression))
        True
        >>> rt=topDownStore.isaBaseQuery("SELECT * { [] a [] }")
        >>> isinstance(rt,(Query,basestring))
        True
        >>> rt=topDownStore.isaBaseQuery("SELECT * { [] a [] OPTIONAL { [] rdfs:seeAlso [] } }")
        >>> isinstance(rt,(BasicGraphPattern,AlgebraExpression))
        True
        """
        from rdflib.sparql.bison.Query import Prolog
        from rdflib.sparql.parser import parse
        global prolog
        prolog = None
        if queryObj:
            query = queryObj
        else:
            query = parse(queryString)
        if not query.prolog:
                query.prolog = Prolog(None, [])
                query.prolog.prefixBindings.update(self.nsBindings)
        else:
            for prefix, nsInst in self.nsBindings.items():
                if prefix not in query.prolog.prefixBindings:
                    query.prolog.prefixBindings[prefix] = nsInst                    
                    
        prolog = query.prolog
        algebra=RenderSPARQLAlgebra(query,nsMappings=self.nsBindings)
        return first(self.getDerivedPredicates(algebra,prolog)) and algebra or query
    
    def __init__(self, 
                store, 
                edb, 
                derivedPredicates=None,
                idb=None,
                DEBUG=False,
                nsBindings={},
                decisionProcedure=TOP_DOWN_METHOD,
                templateMap = None,
                identifyHybridPredicates = False):
        self.dataset = store
        if hasattr(store,'_db'):
            self._db     = store._db
        self.idb               = idb and idb or set()
        self.edb               = edb
        if derivedPredicates is None:
            self.derivedPredicates = list(DerivedPredicateIterator(self.edb,self.idb))
        else:
            self.derivedPredicates = derivedPredicates        
        self.DEBUG             = DEBUG
        self.nsBindings        = nsBindings
        self.decisionProcedure = decisionProcedure
        self.edb.templateMap   = DEFAULT_BUILTIN_MAP if templateMap is None\
                                 else templateMap
        if identifyHybridPredicates:
            self.hybridPredicates = IdentifyHybridPredicates(edb,
                                                             self.derivedPredicates)
        else:
            self.hybridPredicates = None
        
        #Add a cache of the namespace bindings to use later in coining Qnames in 
        #generated queries
        self.edb.revNsMap         = {}
        self.edb.nsMap            = {}
        for k,v in nsBindings.items():
            self.edb.revNsMap[v] = k
            self.edb.nsMap[k]    = v
        for key,uri in self.edb.namespaces():
            self.edb.revNsMap[uri] = key
            self.edb.nsMap[key]    = uri
            
    def invokeDecisionProcedure(self,tp,factGraph,bindings,debug,sipCollection):
        isNotGround = first(itertools.ifilter(lambda i:isinstance(i,Variable),
                                              tp))
        if self.decisionProcedure == TOP_DOWN_METHOD:
            for ans,ns in SipStrategy(
                                 tp,
                                 sipCollection,
                                 factGraph,
                                 self.derivedPredicates,
                                 bindings=bindings,
                                 debug = self.DEBUG,
                                 buildProof = False):
                yield ans,ns
        else:
            rule_store, rule_graph, network = SetupRuleStore(makeNetwork=True)
            bfp = BackwardFixpointProcedure(
                        factGraph,
                        network,
                        self.derivedPredicates,
                        tp,
                        sipCollection,
                        hybridPredicates=self.hybridPredicates,
                        debug=self.DEBUG)
            bfp.createTopDownReteNetwork(self.DEBUG)
            rt=bfp.answers(debug=self.DEBUG)
            if self.DEBUG:
                print >>sys.stderr, "Goal/Query: ", tp
                print >>sys.stderr, "Query was not ground" if isNotGround is not None else "Query was ground"
            if isNotGround is not None:
                for item in bfp.goalSolutions:
                    yield item,None
            else:
                if debug:
                    print >>sys.stderr, bfp.metaInterpNetwork
                    for queryPred in bfp.queryPredicates:
                        if queryPred.arity == 1:
                            bfp.metaInterpNetwork.inferredFacts.remove((None,RDF.type,GetOp(queryPred)))
                        else:
                            bfp.metaInterpNetwork.inferredFacts.remove((None,GetOp(queryPred),None))
                    bfp.metaInterpNetwork.reportConflictSet(True,sys.stderr)
                    yield True,None

    def conjunctiveSipStrategy(self,goalsRemaining,sipCollection,factGraph,bindings={}):
        """
        Given a conjunctive set of triples, invoke sip-strategy passing
        on intermediate solutions to facilitate 'join' behavior
        """
        try:
            tp = goalsRemaining.next()
            assert isinstance(bindings,dict)
            dPred = self.derivedPredicateFromTriple(tp)
            if dPred is None:
                baseEDBQuery = EDBQuery([BuildUnitermFromTuple(tp)],
                                        self.edb,
                                        bindings=bindings)
                if self.DEBUG:
                    print >>sys.stderr,"Evaluating TP against EDB: ",\
                    baseEDBQuery.asSPARQL() 
                query,rt = baseEDBQuery.evaluate()    
                _vars = baseEDBQuery.returnVars
                for item in rt:
                    item.update(bindings)
                    yield item                
            else:
                if self.DEBUG:
                    print >>sys.stderr,"Goal/Query: ", tp
                for nextAnswer,ns in self.invokeDecisionProcedure(
                                            tp,
                                            factGraph,
                                            bindings,
                                            self.DEBUG,
                                            sipCollection):
                    nonGroundGoal = isinstance(nextAnswer,dict) 
                    if nonGroundGoal or nextAnswer:
                        #Either we recieved bindings from top-down evaluation
                        #or we (successfully) proved a ground query
                        if not nonGroundGoal:
                            #Attempt to prove a ground query, return the response
                            rt = nextAnswer
                        else:
                            #Recieved solutions to 'open' query, merge with given bindings
                            #and continue
                            rt = mergeMappings1To2(bindings,nextAnswer)
                        #either answers were provided (the goal wasn't grounded) or
                        #the goal was ground and successfully proved
                        for ansDict in self.conjunctiveSipStrategy(
                                                 goalsRemaining,
                                                 sipCollection,
                                                 factGraph,
                                                 rt):
                            yield ansDict
        except StopIteration:
            yield bindings
            
    def derivedPredicateFromTriple(self,(s,p,o)):
        """
        Given a triple, return its predicate (if derived)
        or None otherwise
        """
        if p in self.derivedPredicates:
            return p
        elif p == RDF.type and o != p and o in self.derivedPredicates:
            return o
        else:
            return None

    def sparql_query(self,
                     queryString,
                     queryObj,
                     graph,
                     dataSetBase,
                     extensionFunctions,
                     initBindings={},
                     initNs={},
                     DEBUG=False):
        """
        The default 'native' SPARQL implementation is based on sparql-p's expansion trees
        layered on top of the read-only RDF APIs of the underlying store 
        """
        from rdflib.sparql.Algebra import TopEvaluate
        from rdflib.QueryResult import QueryResult
        from rdflib import plugin
        from rdflib.sparql.bison.Query import AskQuery
        _expr = self.isaBaseQuery(None,queryObj)
        if isinstance(queryObj.query,AskQuery) and \
           isinstance(_expr,BasicGraphPattern):
            #This is a ground, BGP, involving IDB and can be solved directly
            #using top-down decision procedure
            #First separate out conjunct into EDB and IDB predicates
            #(solving the former first)
            from FuXi.SPARQL import EDBQuery
            groundConjunct  = []
            derivedConjunct = []
            for s,p,o,func in _expr.patterns:
                if self.derivedPredicateFromTriple((s,p,o)) is None:
                    groundConjunct.append(BuildUnitermFromTuple((s,p,o)))
                else:
                    derivedConjunct.append(BuildUnitermFromTuple((s,p,o)))
            if groundConjunct:
                baseEDBQuery = EDBQuery(groundConjunct,self.edb)
                subQuery,ans = baseEDBQuery.evaluate(DEBUG)
                assert isinstance(ans,bool)
            if groundConjunct and not ans:
                askResult = False
            else:
                askResult = True
                for derivedLiteral in derivedConjunct:
                    goal = derivedLiteral.toRDFTuple()
                    #Solve ground, derived goal directly 
                    SetupDDLAndAdornProgram(
                        self.edb,
                        self.idb,
                        [goal],
                        derivedPreds=self.derivedPredicates,
                        ignoreUnboundDPreds = self.decisionProcedure == BFP_METHOD)
                    sipCollection=PrepareSipCollection(self.edb.adornedProgram)
                    rt,node = first(self.invokeDecisionProcedure(
                            goal,
                            self.edb,
                            {},
                            self.DEBUG,
                            sipCollection))
                    if not rt:
                        #Failed to solve goal. If derived predicate is a hybrid 
                        #predicate (or we haven't indicated a distinction), 
                        #check against EDB, if not a hybrid and/or not in EDB
                        #entire ground conjunct fails
                        if (self.hybridPredicates is None or 
                            derivedPred in self.hybridPredicates):
                            if goal not in self.edb:
                                askResult = False
                                break
                        elif self.hybridPredicates is not None:
                            #Purely a IDB predicate, so we fail
                            askResult = False
                            break
            return plugin.get('SPARQLQueryResult',QueryResult)(askResult)
        else:
            rt =   TopEvaluate(queryObj,
                               graph,
                               initBindings,
                               DEBUG=self.DEBUG,
                               dataSetBase=dataSetBase,
                               extensionFunctions=extensionFunctions)
            return plugin.get('SPARQLQueryResult',QueryResult)(rt)
    
    def batch_unify(self, patterns):
        """
        Perform RDF triple store-level unification of a list of triple
        patterns (4-item tuples which correspond to a SPARQL triple pattern
        with an additional constraint for the graph name).  
        
        Uses a SW sip-strategy implementation to solve the conjunctive goal
        and yield unified bindings
        
        :Parameters:
        - `patterns`: a list of 4-item tuples where any of the items can be
           one of: Variable, URIRef, BNode, or Literal.
        
        Returns a generator over dictionaries of solutions to the list of
        triple patterns that are entailed by the regime.    
        """
        dPreds=set()
        goals=[]
        for s,p,o,g in patterns:
            goals.append((s,p,o))
            dPreds.add(p == RDF.type and o or p)
        if set(dPreds).intersection(self.derivedPredicates):
            #Patterns involve derived predicates        
            SetupDDLAndAdornProgram(
                self.edb,
                self.idb,
                goals,
                derivedPreds=self.derivedPredicates,
                ignoreUnboundDPreds = self.decisionProcedure == BFP_METHOD)
    
            sipCollection=PrepareSipCollection(self.edb.adornedProgram)
            if self.DEBUG and sipCollection:
                print >>sys.stderr,"Sideways Information Passing (sip) graph for %s: "%goals
                for sip in SIPRepresentation(sipCollection):
                    print >>sys.stderr,sip
                pprint(list(self.edb.adornedProgram),sys.stderr)
            elif self.DEBUG:
                print >> sys.stderr, "No SIP graph!"
            self.batch_unification = False
            for ansDict in self.conjunctiveSipStrategy(
                                        iter(goals),
                                        sipCollection,
                                        self.edb):
                yield ansDict       
            self.batch_unification = True
        else:
            #conjunctive query involving EDB predicateso only
            vars = []
            triples = []
            for pat in patterns:
                triples.append(BuildUnitermFromTuple(pat[:3]))
                vars.extend([term for term in pat[:3] 
                                if isinstance(term,Variable)])

            query=RDFTuplesToSPARQL(triples,self.edb,vars=vars)
            if self.DEBUG:
                print "Batch unify resolved against EDB"
                print query
            rt = self.edb.query(query,initNs = self.nsBindings)    
            rt = len(vars)>1 and ( dict([(vars[idx],i) 
                                           for idx,i in enumerate(v)]) 
                                                for v in rt ) \
                   or ( dict([(vars[0],v)]) for v in rt )
            for item in rt:
                yield item
            
    def close(self, commit_pending_transaction=False):
        """
        This closes the database connection. The commit_pending_transaction parameter specifies whether to
        commit all pending transactions before closing (if the store is transactional).
        """
        return self.dataset.close(commit_pending_transaction)
    
    def destroy(self, configuration):
        """
        This destroys the instance of the store identified by the configuration string.
        """
        return self.dataset.destroy(configuration)
        
    def triples_choices(self, (subject, predicate, object_),context=None):
        """
        A variant of triples that can take a list of terms instead of a single
        term in any slot.  Stores can implement this to optimize the response time
        from the default 'fallback' implementation, which will iterate
        over each term in the list and dispatch to tripless
        """
        for rt in self.dataset.triples_choices((subject, predicate, object_),context):
            yield rt

    def triples(self, (subject, predicate, object), context=None):
        """
        A generator over all the triples matching the pattern. Pattern can
        include any objects for used for comparing against nodes in the store, for
        example, REGEXTerm, URIRef, Literal, BNode, Variable, Graph, QuotedGraph, Date? DateRange?

        A conjunctive query can be indicated by either providing a value of None
        for the context or the identifier associated with the Conjunctive Graph (if it's context aware).
        """
        for rt in self.dataset.triples((subject, predicate, object), context):
            yield rt

    def __len__(self, context=None):
        """
        Number of statements in the store. This should only account for non-quoted (asserted) statements
        if the context is not specified, otherwise it should return the number of statements in the formula or context given.
        """
        return len(self.dataset)

    def contexts(self, triple=None):
        """
        Generator over all contexts in the graph. If triple is specified, a generator over all
        contexts the triple is in.
        """
        for ctx in self.dataset.contexts(triple):
            yield ctx

    # Optional Namespace methods

    def bind(self, prefix, namespace):
        self.nsBindings[prefix]=namespace
        #self.targetGraph.bind(prefix,namespace)

    def prefix(self, namespace):
        revDict = dict([(v,k) for k,v in self.nsBindings.items()])
        return revDict.get(namespace)

    def namespace(self, prefix):
        return self.nsBindings.get(prefix)

    def namespaces(self):
        for prefix,nsUri in self.nsBindings.items():
            yield prefix,nsUri

    # Optional Transactional methods

    def commit(self):
        self.dataset.commit()

    def rollback(self):
        self.dataset.rollback()
    
def test():
     import doctest
     doctest.testmod()

if __name__ == '__main__':
    test()


#!/usr/bin/env python
# encoding: utf-8
import unittest
from rdflib.Graph import Graph
from FuXi.Rete.RuleStore import SetupRuleStore

## fix for bug in reset method which didn't initialise
## network.inferredFacts properly if the provided graph
## was empty
## http://code.google.com/p/fuxi/issues/detail?id=17
##

class NetworkReset(unittest.TestCase):
    def setUp(self):
	    self.rule_store, self.rule_graph, self.network = SetupRuleStore(makeNetwork=True)
    def testReset(self):
        newInferredFacts = Graph()
        self.network.reset(newInferredFacts)
        self.failUnless(newInferredFacts is self.network.inferredFacts)		
    
if __name__ == '__main__':
	unittest.main()


import unittest, os, time, sys 
from cStringIO import StringIO 
from rdflib import RDF, URIRef 
from FuXi.Rete import * 
from FuXi.Rete.RuleStore import N3RuleStore, SetupRuleStore 
from FuXi.Rete.Util import renderNetwork, generateTokenSet 
from FuXi.Horn.PositiveConditions import Uniterm, BuildUnitermFromTuple 
from FuXi.Horn.HornRules import HornFromN3 
from rdflib import plugin, Namespace 
from rdflib.store import Store 
from rdflib.Graph import Graph 
N3_PROGRAM=\
""" 
@prefix m: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
{ ?det a m:Detection. 
  ?det has m:name ?infName. 
} => { 

  ?det has m:inference [ a m:Inference; m:inference_name ?infName ]. 
}. 

""" 
N3_FACTS=\
""" 
@prefix m: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
m:Detection a rdfs:Class. 
m:Inference a rdfs:Class. 
:det1 a m:Detection. 
:det1 m:name "Inference1". 
:det2 a m:Detection. 
:det2 m:name "Inference2". 
""" 
class ExistentialInHeadTest(unittest.TestCase): 
    def testExistentials(self): 
        store = plugin.get('IOMemory',Store)() 
        store.open('') 
        ruleStore = N3RuleStore() 
        ruleGraph = Graph(ruleStore) 
        ruleGraph.parse(StringIO(N3_PROGRAM),format='n3') 
        factGraph = Graph(store) 
        factGraph.parse(StringIO(N3_FACTS),format='n3') 
        deltaGraph = Graph(store) 
        network = ReteNetwork(ruleStore, 
                              initialWorkingMemory=generateTokenSet(factGraph), 
                              inferredTarget = deltaGraph) 
        inferenceCount = 0 
        for inferredFact in network.inferredFacts.subjects(
                                       predicate=RDF.type, 
                                       object=URIRef('http://example.com/#Inference')): 
            inferenceCount = inferenceCount + 1
        print network.inferredFacts.serialize(format='n3') 
        self.failUnless(inferenceCount > 1,  'Each rule firing should introduce a new BNode!') 
        cg = network.closureGraph(factGraph, store=ruleStore) 
        #print cg.serialize(format="n3") 

SKOLEM_MACHINE_RULES=\
""" 
@prefix ex: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
{?X ex:b ?Y} => {_:Z ex:p ?Y}. 
{?X ex:e ?Y} => {_:Z ex:p ?Y}. 
""" 
SKOLEM_MACHINE_FACTS=\
""" 
@prefix ex: <http://example.com/#>. 
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> . 
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> . 
ex:a ex:b ex:c. 
ex:d ex:e ex:c. 
""" 

#{?x ?p ?y} => {[] ?p []} . :a :b :c . 

EX_NS = Namespace('http://example.com/#')
             
class SkolemMachine(unittest.TestCase):
    def setUp(self):
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        self.network= network
        self.factGraph = Graph().parse(StringIO(SKOLEM_MACHINE_FACTS),format='n3')
        for rule in HornFromN3(StringIO(SKOLEM_MACHINE_RULES)):
            self.network.buildNetworkFromClause(rule)
        self.network.feedFactsToAdd(generateTokenSet(self.factGraph))
#    def testSkolemMachine(self):
#        self.assertEquals(len(list(self.network.inferredFacts.triples(
#                                          (None,EX_NS.p,None)))),
#                          1)
            
if __name__ == "__main__": 
    unittest.main() 